from collections.abc import Iterable, Iterator
from logging import Logger
from typing import Any
from typing_extensions import Literal

chardet_type: Any

def chardet_dammit(s): ...

xml_encoding: str
html_meta: str
encoding_res: Any

class EntitySubstitution:
    CHARACTER_TO_HTML_ENTITY: Any
    HTML_ENTITY_TO_CHARACTER: Any
    CHARACTER_TO_HTML_ENTITY_RE: Any
    CHARACTER_TO_XML_ENTITY: Any
    BARE_AMPERSAND_OR_BRACKET: Any
    AMPERSAND_OR_BRACKET: Any
    @classmethod
    def quoted_attribute_value(cls, value): ...
    @classmethod
    def substitute_xml(cls, value, make_quoted_attribute: bool = ...): ...
    @classmethod
    def substitute_xml_containing_entities(cls, value, make_quoted_attribute: bool = ...): ...
    @classmethod
    def substitute_html(cls, s): ...

class EncodingDetector:
    known_definite_encodings: list[str]
    user_encodings: list[str]
    exclude_encodings: set[str]
    chardet_encoding: Any | None
    is_html: bool
    declared_encoding: str | None
    markup: Any
    sniffed_encoding: str | None
    def __init__(
        self,
        markup,
        known_definite_encodings: Iterable[str] | None = ...,
        is_html: bool = ...,
        exclude_encodings: list[str] | None = ...,
        user_encodings: list[str] | None = ...,
        override_encodings: list[str] | None = ...,
    ) -> None: ...
    @property
    def encodings(self) -> Iterator[str]: ...
    @classmethod
    def strip_byte_order_mark(cls, data): ...
    @classmethod
    def find_declared_encoding(cls, markup, is_html: bool = ..., search_entire_document: bool = ...) -> str | None: ...

class UnicodeDammit:
    CHARSET_ALIASES: dict[str, str]
    ENCODINGS_WITH_SMART_QUOTES: list[str]
    smart_quotes_to: Literal["ascii", "xml", "html"] | None
    tried_encodings: list[tuple[str, str]]
    contains_replacement_characters: bool
    is_html: bool
    log: Logger
    detector: EncodingDetector
    markup: Any
    unicode_markup: str
    original_encoding: Any | None
    def __init__(
        self,
        markup,
        known_definite_encodings: list[str] | None = ...,
        smart_quotes_to: Literal["ascii", "xml", "html"] | None = ...,
        is_html: bool = ...,
        exclude_encodings: list[str] | None = ...,
        user_encodings: list[str] | None = ...,
        override_encodings: list[str] | None = ...,
    ) -> None: ...
    @property
    def declared_html_encoding(self) -> str | None: ...
    def find_codec(self, charset: str) -> str | None: ...
    MS_CHARS: dict[bytes, str | tuple[str, ...]]
    MS_CHARS_TO_ASCII: dict[bytes, str]
    WINDOWS_1252_TO_UTF8: dict[int, bytes]
    MULTIBYTE_MARKERS_AND_SIZES: list[tuple[int, int, int]]
    FIRST_MULTIBYTE_MARKER: int
    LAST_MULTIBYTE_MARKER: int
    @classmethod
    def detwingle(cls, in_bytes: bytes, main_encoding: str = ..., embedded_encoding: str = ...) -> bytes: ...



"""Base visitor for typed_ast parse trees."""

import sys

from pytype.ast import visitor as ast_visitor
from pytype.pyi import types
from pytype.pyi.types import ParseError  # pylint: disable=g-importing-member

# pylint: disable=g-import-not-at-top
if sys.version_info >= (3, 8):
  import ast as ast3
else:
  from typed_ast import ast3
# pylint: enable=g-import-not-at-top


class BaseVisitor(ast_visitor.BaseVisitor):
  """Base visitor for all typed_ast visitors.

  - Reraises ParseError with position information.
  - Handles literal constants
  - Has an optional Definitions member
  """

  def __init__(self, *, defs=None, filename=None):
    super().__init__(ast3, visit_decorators=False)
    self.defs = defs
    self.filename = filename  # used for error messages
    self.src_code = None  # set in subclass, used for error messages
    # Keep track of the name being subscripted. See AnnotationVisitor.visit_Name
    # for why this is needed.
    self.subscripted = []

  def _call_visitor(self, node):
    try:
      return super()._call_visitor(node)
    except Exception as e:  # pylint: disable=broad-except
      raise ParseError.from_exc(e).at(node, self.filename, self.src_code)

  def enter(self, node):
    try:
      return super().enter(node)
    except Exception as e:  # pylint: disable=broad-except
      raise ParseError.from_exc(e).at(node, self.filename, self.src_code)

  def leave(self, node):
    try:
      return super().leave(node)
    except Exception as e:  # pylint: disable=broad-except
      raise ParseError.from_exc(e).at(node, self.filename, self.src_code)

  def visit_Ellipsis(self, node):
    return self.defs.ELLIPSIS

  def visit_Constant(self, node):
    if node.value is Ellipsis:
      return self.defs.ELLIPSIS
    return types.Pyval.from_const(node)

  def visit_NameConstant(self, node):
    return types.Pyval.from_const(node)

  def visit_Num(self, node):
    return types.Pyval.from_num(node)

  def visit_Str(self, node):
    return types.Pyval.from_str(node)

  def visit_Bytes(self, node):
    return self.visit_Str(node)

  def visit_UnaryOp(self, node):
    if isinstance(node.op, ast3.USub):
      if isinstance(node.operand, types.Pyval):
        return node.operand.negated()
    raise ParseError(f"Unexpected unary operator: {node.op}")



from typing import Any

from braintree.attribute_getter import AttributeGetter as AttributeGetter

class AddressDetails(AttributeGetter):
    detail_list: Any
    def __init__(self, attributes) -> None: ...



def f1(x):
    pass

def f2(x):
    pass

def f3(x):
    pass

def f4(x):
    pass

def f5(x):
    pass

def f6(x):
    pass

def f7(x):
    pass

def f8(x):
    pass

def f9(x):
    pass

def f10(x):
    pass



import subprocess
import sys
from _typeshed import Self
from types import TracebackType
from typing import Any, AnyStr, Callable, Protocol
from typing_extensions import Literal

if sys.platform == "win32":
    if sys.version_info >= (3, 7):
        __all__ = ("pipe", "Popen", "PIPE", "PipeHandle")
    else:
        __all__ = ["socketpair", "pipe", "Popen", "PIPE", "PipeHandle"]
        import socket

        socketpair = socket.socketpair

    class _WarnFunction(Protocol):
        def __call__(
            self, message: str, category: type[Warning] = ..., stacklevel: int = ..., source: PipeHandle = ...
        ) -> None: ...
    BUFSIZE: Literal[8192]
    PIPE = subprocess.PIPE
    STDOUT = subprocess.STDOUT
    def pipe(*, duplex: bool = ..., overlapped: tuple[bool, bool] = ..., bufsize: int = ...) -> tuple[int, int]: ...

    class PipeHandle:
        def __init__(self, handle: int) -> None: ...
        if sys.version_info >= (3, 8):
            def __del__(self, _warn: _WarnFunction = ...) -> None: ...
        else:
            def __del__(self) -> None: ...

        def __enter__(self: Self) -> Self: ...
        def __exit__(self, t: type[BaseException] | None, v: BaseException | None, tb: TracebackType | None) -> None: ...
        @property
        def handle(self) -> int: ...
        def fileno(self) -> int: ...
        def close(self, *, CloseHandle: Callable[[int], None] = ...) -> None: ...

    class Popen(subprocess.Popen[AnyStr]):
        stdin: PipeHandle | None  # type: ignore[assignment]
        stdout: PipeHandle | None  # type: ignore[assignment]
        stderr: PipeHandle | None  # type: ignore[assignment]
        # For simplicity we omit the full overloaded __new__ signature of
        # subprocess.Popen. The arguments are mostly the same, but
        # subprocess.Popen takes other positional-or-keyword arguments before
        # stdin.
        def __new__(
            cls: type[Self],
            args: subprocess._CMD,
            stdin: subprocess._FILE | None = ...,
            stdout: subprocess._FILE | None = ...,
            stderr: subprocess._FILE | None = ...,
            **kwds: Any,
        ) -> Self: ...
        def __init__(
            self,
            args: subprocess._CMD,
            stdin: subprocess._FILE | None = ...,
            stdout: subprocess._FILE | None = ...,
            stderr: subprocess._FILE | None = ...,
            **kwds: Any,
        ) -> None: ...



from types import TracebackType
from typing import IO, Any, Callable, ContextManager, Iterable, Iterator, Protocol, TypeVar
from typing_extensions import ParamSpec

_T = TypeVar("_T")
_T_co = TypeVar("_T_co", covariant=True)
_F = TypeVar("_F", bound=Callable[..., Any])
_P = ParamSpec("_P")

_ExitFunc = Callable[[type[BaseException] | None, BaseException | None, TracebackType | None], bool]

class GeneratorContextManager(ContextManager[_T_co]):
    def __call__(self, func: _F) -> _F: ...

def contextmanager(func: Callable[_P, Iterator[_T]]) -> Callable[_P, ContextManager[_T]]: ...
def nested(*mgr: ContextManager[Any]) -> ContextManager[Iterable[Any]]: ...

class _SupportsClose(Protocol):
    def close(self) -> None: ...

_SupportsCloseT = TypeVar("_SupportsCloseT", bound=_SupportsClose)

class closing(ContextManager[_SupportsCloseT]):
    def __init__(self, thing: _SupportsCloseT) -> None: ...



from typing import Any

class ReadOnlyCell:
    parent: Any
    row: Any
    column: Any
    data_type: Any
    def __init__(self, sheet, row, column, value, data_type: str = ..., style_id: int = ...) -> None: ...
    def __eq__(self, other): ...
    def __ne__(self, other): ...
    # defined twice in the implementation
    @property
    def coordinate(self): ...
    @property
    def column_letter(self): ...
    @property
    def style_array(self): ...
    @property
    def has_style(self): ...
    @property
    def number_format(self): ...
    @property
    def font(self): ...
    @property
    def fill(self): ...
    @property
    def border(self): ...
    @property
    def alignment(self): ...
    @property
    def protection(self): ...
    @property
    def is_date(self): ...
    @property
    def internal_value(self): ...
    @property
    def value(self): ...
    @value.setter
    def value(self, value) -> None: ...

class EmptyCell:
    value: Any
    is_date: bool
    font: Any
    border: Any
    fill: Any
    number_format: Any
    alignment: Any
    data_type: str

EMPTY_CELL: Any



from typing import Any, NamedTuple, Union
from typing_extensions import Literal

class _TagInfo(NamedTuple):
    value: Any
    name: str
    type: _TagType
    length: int
    enum: dict[str, int]

class TagInfo(_TagInfo):
    def __new__(
        cls,
        value: Any | None = ...,
        name: str = ...,
        type: _TagType | None = ...,
        length: int | None = ...,
        enum: dict[str, int] | None = ...,
    ): ...
    def cvt_enum(self, value): ...

def lookup(tag: int, group: int | None = ...) -> _TagInfo: ...

BYTE: Literal[1]
ASCII: Literal[2]
SHORT: Literal[3]
LONG: Literal[4]
RATIONAL: Literal[5]
SIGNED_BYTE: Literal[6]
UNDEFINED: Literal[7]
SIGNED_SHORT: Literal[8]
SIGNED_LONG: Literal[9]
SIGNED_RATIONAL: Literal[10]
FLOAT: Literal[11]
DOUBLE: Literal[12]
IFD: Literal[13]

_TagType = Literal[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13]
_TagTuple = Union[tuple[str, _TagType, int], tuple[str, _TagInfo, int, dict[str, int]]]

TAGS_V2: dict[int, _TagTuple]
TAGS_V2_GROUPS: dict[int, dict[int, _TagTuple]]
TAGS: dict[int, str]
TYPES: dict[int, str]
LIBTIFF_CORE: set[int]



from abc import ABCMeta, abstractmethod
from typing import ClassVar

from cryptography.hazmat.primitives.hashes import HashAlgorithm

class AsymmetricPadding(metaclass=ABCMeta):
    @property
    @abstractmethod
    def name(self) -> str: ...

class MGF1:
    def __init__(self, algorithm: HashAlgorithm) -> None: ...

class OAEP(AsymmetricPadding):
    def __init__(self, mgf: MGF1, algorithm: HashAlgorithm, label: bytes | None) -> None: ...
    @property
    def name(self) -> str: ...

class PKCS1v15(AsymmetricPadding):
    @property
    def name(self) -> str: ...

class PSS(AsymmetricPadding):
    MAX_LENGTH: ClassVar[object]
    def __init__(self, mgf: MGF1, salt_length: int | object) -> None: ...
    @property
    def name(self) -> str: ...



class DefaultStreaming:
    def __init__(self, streaming_threshold: int = ...) -> None: ...
    def is_eligible(self, segment): ...
    def stream(self, entity, callback) -> None: ...
    @property
    def streaming_threshold(self): ...
    @streaming_threshold.setter
    def streaming_threshold(self, value) -> None: ...



from _typeshed import BytesPath, StrOrBytesPath, StrPath
from genericpath import (
    commonprefix as commonprefix,
    exists as exists,
    getatime as getatime,
    getctime as getctime,
    getmtime as getmtime,
    getsize as getsize,
    isdir as isdir,
    isfile as isfile,
    samefile as samefile,
    sameopenfile as sameopenfile,
    samestat as samestat,
)
from os import PathLike

# Re-export common definitions from posixpath to reduce duplication
from posixpath import (
    abspath as abspath,
    curdir as curdir,
    defpath as defpath,
    devnull as devnull,
    expanduser as expanduser,
    expandvars as expandvars,
    extsep as extsep,
    isabs as isabs,
    lexists as lexists,
    pardir as pardir,
    pathsep as pathsep,
    sep as sep,
    splitdrive as splitdrive,
    splitext as splitext,
    supports_unicode_filenames as supports_unicode_filenames,
)
from typing import AnyStr, overload

altsep: str | None

@overload
def basename(s: PathLike[AnyStr]) -> AnyStr: ...
@overload
def basename(s: AnyStr) -> AnyStr: ...
@overload
def dirname(s: PathLike[AnyStr]) -> AnyStr: ...
@overload
def dirname(s: AnyStr) -> AnyStr: ...
@overload
def normcase(path: PathLike[AnyStr]) -> AnyStr: ...
@overload
def normcase(path: AnyStr) -> AnyStr: ...
@overload
def normpath(s: PathLike[AnyStr]) -> AnyStr: ...
@overload
def normpath(s: AnyStr) -> AnyStr: ...
@overload
def realpath(path: PathLike[AnyStr]) -> AnyStr: ...
@overload
def realpath(path: AnyStr) -> AnyStr: ...
def islink(s: StrOrBytesPath) -> bool: ...

# Mypy complains that the signatures overlap, but things seem to behave correctly anyway.
@overload
def join(s: StrPath, *paths: StrPath) -> str: ...
@overload
def join(s: BytesPath, *paths: BytesPath) -> bytes: ...
@overload
def split(s: PathLike[AnyStr]) -> tuple[AnyStr, AnyStr]: ...
@overload
def split(s: AnyStr) -> tuple[AnyStr, AnyStr]: ...



import sys
from typing import Any, AnyStr, Callable, Generic, Mapping, NamedTuple, Sequence, overload

if sys.version_info >= (3, 9):
    from types import GenericAlias

__all__ = [
    "urlparse",
    "urlunparse",
    "urljoin",
    "urldefrag",
    "urlsplit",
    "urlunsplit",
    "urlencode",
    "parse_qs",
    "parse_qsl",
    "quote",
    "quote_plus",
    "quote_from_bytes",
    "unquote",
    "unquote_plus",
    "unquote_to_bytes",
    "DefragResult",
    "ParseResult",
    "SplitResult",
    "DefragResultBytes",
    "ParseResultBytes",
    "SplitResultBytes",
]

_Str = bytes | str

uses_relative: list[str]
uses_netloc: list[str]
uses_params: list[str]
non_hierarchical: list[str]
uses_query: list[str]
uses_fragment: list[str]
scheme_chars: str
MAX_CACHE_SIZE: int

class _ResultMixinBase(Generic[AnyStr]):
    def geturl(self) -> AnyStr: ...

class _ResultMixinStr(_ResultMixinBase[str]):
    def encode(self, encoding: str = ..., errors: str = ...) -> _ResultMixinBytes: ...

class _ResultMixinBytes(_ResultMixinBase[bytes]):
    def decode(self, encoding: str = ..., errors: str = ...) -> _ResultMixinStr: ...

class _NetlocResultMixinBase(Generic[AnyStr]):
    @property
    def username(self) -> AnyStr | None: ...
    @property
    def password(self) -> AnyStr | None: ...
    @property
    def hostname(self) -> AnyStr | None: ...
    @property
    def port(self) -> int | None: ...
    if sys.version_info >= (3, 9):
        def __class_getitem__(cls, item: Any) -> GenericAlias: ...

class _NetlocResultMixinStr(_NetlocResultMixinBase[str], _ResultMixinStr): ...
class _NetlocResultMixinBytes(_NetlocResultMixinBase[bytes], _ResultMixinBytes): ...

class _DefragResultBase(tuple[Any, ...], Generic[AnyStr]):
    if sys.version_info >= (3, 10):
        __match_args__ = ("url", "fragment")
    @property
    def url(self) -> AnyStr: ...
    @property
    def fragment(self) -> AnyStr: ...

class _SplitResultBase(NamedTuple):
    scheme: str
    netloc: str
    path: str
    query: str
    fragment: str

class _SplitResultBytesBase(NamedTuple):
    scheme: bytes
    netloc: bytes
    path: bytes
    query: bytes
    fragment: bytes

class _ParseResultBase(NamedTuple):
    scheme: str
    netloc: str
    path: str
    params: str
    query: str
    fragment: str

class _ParseResultBytesBase(NamedTuple):
    scheme: bytes
    netloc: bytes
    path: bytes
    params: bytes
    query: bytes
    fragment: bytes

# Structured result objects for string data
class DefragResult(_DefragResultBase[str], _ResultMixinStr): ...
class SplitResult(_SplitResultBase, _NetlocResultMixinStr): ...
class ParseResult(_ParseResultBase, _NetlocResultMixinStr): ...

# Structured result objects for bytes data
class DefragResultBytes(_DefragResultBase[bytes], _ResultMixinBytes): ...
class SplitResultBytes(_SplitResultBytesBase, _NetlocResultMixinBytes): ...
class ParseResultBytes(_ParseResultBytesBase, _NetlocResultMixinBytes): ...

def parse_qs(
    qs: AnyStr | None,
    keep_blank_values: bool = ...,
    strict_parsing: bool = ...,
    encoding: str = ...,
    errors: str = ...,
    max_num_fields: int | None = ...,
    separator: str = ...,
) -> dict[AnyStr, list[AnyStr]]: ...
def parse_qsl(
    qs: AnyStr | None,
    keep_blank_values: bool = ...,
    strict_parsing: bool = ...,
    encoding: str = ...,
    errors: str = ...,
    max_num_fields: int | None = ...,
    separator: str = ...,
) -> list[tuple[AnyStr, AnyStr]]: ...
@overload
def quote(string: str, safe: _Str = ..., encoding: str | None = ..., errors: str | None = ...) -> str: ...
@overload
def quote(string: bytes, safe: _Str = ...) -> str: ...
def quote_from_bytes(bs: bytes, safe: _Str = ...) -> str: ...
@overload
def quote_plus(string: str, safe: _Str = ..., encoding: str | None = ..., errors: str | None = ...) -> str: ...
@overload
def quote_plus(string: bytes, safe: _Str = ...) -> str: ...
def unquote(string: str, encoding: str = ..., errors: str = ...) -> str: ...
def unquote_to_bytes(string: _Str) -> bytes: ...
def unquote_plus(string: str, encoding: str = ..., errors: str = ...) -> str: ...
@overload
def urldefrag(url: str) -> DefragResult: ...
@overload
def urldefrag(url: bytes | None) -> DefragResultBytes: ...
def urlencode(
    query: Mapping[Any, Any] | Mapping[Any, Sequence[Any]] | Sequence[tuple[Any, Any]] | Sequence[tuple[Any, Sequence[Any]]],
    doseq: bool = ...,
    safe: _Str = ...,
    encoding: str = ...,
    errors: str = ...,
    quote_via: Callable[[AnyStr, _Str, str, str], str] = ...,
) -> str: ...
def urljoin(base: AnyStr, url: AnyStr | None, allow_fragments: bool = ...) -> AnyStr: ...
@overload
def urlparse(url: str, scheme: str | None = ..., allow_fragments: bool = ...) -> ParseResult: ...
@overload
def urlparse(url: bytes | None, scheme: bytes | None = ..., allow_fragments: bool = ...) -> ParseResultBytes: ...
@overload
def urlsplit(url: str, scheme: str | None = ..., allow_fragments: bool = ...) -> SplitResult: ...
@overload
def urlsplit(url: bytes | None, scheme: bytes | None = ..., allow_fragments: bool = ...) -> SplitResultBytes: ...
@overload
def urlunparse(
    components: tuple[AnyStr | None, AnyStr | None, AnyStr | None, AnyStr | None, AnyStr | None, AnyStr | None]
) -> AnyStr: ...
@overload
def urlunparse(components: Sequence[AnyStr | None]) -> AnyStr: ...
@overload
def urlunsplit(components: tuple[AnyStr | None, AnyStr | None, AnyStr | None, AnyStr | None, AnyStr | None]) -> AnyStr: ...
@overload
def urlunsplit(components: Sequence[AnyStr | None]) -> AnyStr: ...
def unwrap(url: str) -> str: ...



def f1(x: e1="foo") -> r1:
    pass

def f2(x: e2=
    #
"foo") -> r2:
    pass

def f3(#
 x: e3#=
  = #
    #
 123#
#
) -> r3:
    pass


def f4(x: e4=(1,2)) -> r4:
    pass

def f5(x: e5=(1,)) -> r5:
    pass

def f6(x: e6=int) -> r6:
    pass

# static analysis would give error here
def f7(x : int=int):
    pass

def f8(x: e8={1:2}) -> r8:
    pass

def f9(x: e9=[1,2][:1]) -> r9:
    pass



from typing import Any, Iterable, Match, Pattern as _Pattern

SPECIAL_CHARS: str
REPEAT_CHARS: str
DIGITS: set[Any]
OCTDIGITS: set[Any]
HEXDIGITS: set[Any]
WHITESPACE: set[Any]
ESCAPES: dict[str, tuple[str, int]]
CATEGORIES: dict[str, tuple[str, str] | tuple[str, list[tuple[str, str]]]]
FLAGS: dict[str, int]

class Pattern:
    flags: int
    open: list[int]
    groups: int
    groupdict: dict[str, int]
    lookbehind: int
    def __init__(self) -> None: ...
    def opengroup(self, name: str = ...) -> int: ...
    def closegroup(self, gid: int) -> None: ...
    def checkgroup(self, gid: int) -> bool: ...

_OpSubpatternType = tuple[int | None, int, int, SubPattern]
_OpGroupRefExistsType = tuple[int, SubPattern, SubPattern]
_OpInType = list[tuple[str, int]]
_OpBranchType = tuple[None, list[SubPattern]]
_AvType = _OpInType | _OpBranchType | Iterable[SubPattern] | _OpGroupRefExistsType | _OpSubpatternType
_CodeType = str | _AvType

class SubPattern:
    pattern: str
    data: list[_CodeType]
    width: int | None
    def __init__(self, pattern, data: list[_CodeType] = ...) -> None: ...
    def dump(self, level: int = ...) -> None: ...
    def __len__(self) -> int: ...
    def __delitem__(self, index: int | slice) -> None: ...
    def __getitem__(self, index: int | slice) -> SubPattern | _CodeType: ...
    def __setitem__(self, index: int | slice, code: _CodeType): ...
    def insert(self, index, code: _CodeType) -> None: ...
    def append(self, code: _CodeType) -> None: ...
    def getwidth(self) -> int: ...

class Tokenizer:
    string: str
    index: int
    def __init__(self, string: str) -> None: ...
    def match(self, char: str, skip: int = ...) -> int: ...
    def get(self) -> str | None: ...
    def tell(self) -> tuple[int, str | None]: ...
    def seek(self, index: int) -> None: ...

def isident(char: str) -> bool: ...
def isdigit(char: str) -> bool: ...
def isname(name: str) -> bool: ...
def parse(str: str, flags: int = ..., pattern: Pattern = ...) -> SubPattern: ...

_Template = tuple[list[tuple[int, int]], list[int | None]]

def parse_template(source: str, pattern: _Pattern[Any]) -> _Template: ...
def expand_template(template: _Template, match: Match[Any]) -> str: ...



import socket
from typing import Any, BinaryIO, Pattern, Text, overload

_LongResp = tuple[bytes, list[bytes], int]

class error_proto(Exception): ...

POP3_PORT: int
POP3_SSL_PORT: int
CR: bytes
LF: bytes
CRLF: bytes

class POP3:
    host: Text
    port: int
    sock: socket.socket
    file: BinaryIO
    welcome: bytes
    def __init__(self, host: Text, port: int = ..., timeout: float = ...) -> None: ...
    def getwelcome(self) -> bytes: ...
    def set_debuglevel(self, level: int) -> None: ...
    def user(self, user: Text) -> bytes: ...
    def pass_(self, pswd: Text) -> bytes: ...
    def stat(self) -> tuple[int, int]: ...
    def list(self, which: Any | None = ...) -> _LongResp: ...
    def retr(self, which: Any) -> _LongResp: ...
    def dele(self, which: Any) -> bytes: ...
    def noop(self) -> bytes: ...
    def rset(self) -> bytes: ...
    def quit(self) -> bytes: ...
    def close(self) -> None: ...
    def rpop(self, user: Text) -> bytes: ...
    timestamp: Pattern[Text]
    def apop(self, user: Text, secret: Text) -> bytes: ...
    def top(self, which: Any, howmuch: int) -> _LongResp: ...
    @overload
    def uidl(self) -> _LongResp: ...
    @overload
    def uidl(self, which: Any) -> bytes: ...

class POP3_SSL(POP3):
    def __init__(
        self, host: Text, port: int = ..., keyfile: Text | None = ..., certfile: Text | None = ..., timeout: float = ...
    ) -> None: ...



from typing import Any

from selenium.webdriver.chrome.webdriver import WebDriver as ChromiumDriver

from .options import Options as Options

class OperaDriver(ChromiumDriver):
    def __init__(
        self,
        executable_path: Any | None = ...,
        port: int = ...,
        options: Any | None = ...,
        service_args: Any | None = ...,
        desired_capabilities: Any | None = ...,
        service_log_path: Any | None = ...,
        opera_options: Any | None = ...,
        keep_alive: bool = ...,
    ) -> None: ...
    def create_options(self): ...

class WebDriver(OperaDriver):
    class ServiceType:
        CHROMIUM: int
    def __init__(
        self,
        desired_capabilities: Any | None = ...,
        executable_path: Any | None = ...,
        port: int = ...,
        service_log_path: Any | None = ...,
        service_args: Any | None = ...,
        options: Any | None = ...,
    ) -> None: ...



"""Tests for directors.py."""

import sys
import textwrap

from pytype import blocks
from pytype import errors
from pytype.directors import directors
from pytype.pyc import pyc
import unittest

_TEST_FILENAME = "my_file.py"


class LineSetTest(unittest.TestCase):

  def test_no_ranges(self):
    lines = directors._LineSet()
    lines.set_line(2, True)
    self.assertNotIn(0, lines)
    self.assertNotIn(1, lines)
    self.assertIn(2, lines)
    self.assertNotIn(3, lines)

  def test_closed_range(self):
    lines = directors._LineSet()
    lines.start_range(2, True)
    lines.start_range(4, False)
    self.assertNotIn(1, lines)
    self.assertIn(2, lines)
    self.assertIn(3, lines)
    self.assertNotIn(4, lines)
    self.assertNotIn(1000, lines)

  def test_open_range(self):
    lines = directors._LineSet()
    lines.start_range(2, True)
    lines.start_range(4, False)
    lines.start_range(7, True)
    self.assertNotIn(1, lines)
    self.assertIn(2, lines)
    self.assertIn(3, lines)
    self.assertNotIn(4, lines)
    self.assertNotIn(5, lines)
    self.assertNotIn(6, lines)
    self.assertIn(7, lines)
    self.assertIn(1000, lines)

  def test_range_at_zero(self):
    lines = directors._LineSet()
    lines.start_range(0, True)
    lines.start_range(3, False)
    self.assertNotIn(-1, lines)
    self.assertIn(0, lines)
    self.assertIn(1, lines)
    self.assertIn(2, lines)
    self.assertNotIn(3, lines)

  def test_line_overrides_range(self):
    lines = directors._LineSet()
    lines.start_range(2, True)
    lines.start_range(5, False)
    lines.set_line(3, False)
    self.assertIn(2, lines)
    self.assertNotIn(3, lines)
    self.assertIn(4, lines)

  def test_redundant_range(self):
    lines = directors._LineSet()
    lines.start_range(2, True)
    lines.start_range(3, True)
    lines.start_range(5, False)
    lines.start_range(9, False)
    self.assertNotIn(1, lines)
    self.assertIn(2, lines)
    self.assertIn(3, lines)
    self.assertIn(4, lines)
    self.assertNotIn(5, lines)
    self.assertNotIn(9, lines)
    self.assertNotIn(1000, lines)

  def test_enable_disable_on_same_line(self):
    lines = directors._LineSet()
    lines.start_range(2, True)
    lines.start_range(2, False)
    lines.start_range(3, True)
    lines.start_range(5, False)
    lines.start_range(5, True)
    self.assertNotIn(2, lines)
    self.assertIn(3, lines)
    self.assertIn(4, lines)
    self.assertIn(5, lines)
    self.assertIn(1000, lines)

  def test_decreasing_lines_not_allowed(self):
    lines = directors._LineSet()
    self.assertRaises(ValueError, lines.start_range, -100, True)
    lines.start_range(2, True)
    self.assertRaises(ValueError, lines.start_range, 1, True)


class DirectorTestCase(unittest.TestCase):

  python_version = sys.version_info[:2]

  @classmethod
  def setUpClass(cls):
    super().setUpClass()
    # Invoking the _error_name decorator will register the name as a valid
    # error name.
    for name in ["test-error", "test-other-error"]:
      errors._error_name(name)

  def _create(self, src, disable=()):
    self.num_lines = len(src.rstrip().splitlines())
    src = textwrap.dedent(src)
    src_tree = directors.parse_src(src, self.python_version)
    self._errorlog = errors.ErrorLog()
    if self.python_version < (3, 8):
      raw_code = pyc.compile_src(src, _TEST_FILENAME, self.python_version, None)
      code = blocks.process_code(raw_code, self.python_version)
    else:
      code = None
    self._director = directors.Director(
        src_tree, self._errorlog, _TEST_FILENAME, disable, code)

  def _should_report(self, expected, lineno, error_name="test-error",
                     filename=_TEST_FILENAME):
    error = errors.Error.for_test(
        errors.SEVERITY_ERROR, "message", error_name, filename=filename,
        lineno=lineno)
    self.assertEqual(
        expected,
        self._director.filter_error(error))


class DirectorTest(DirectorTestCase):

  def test_ignore_globally(self):
    self._create("", ["my-error"])
    self._should_report(False, 42, error_name="my-error")

  def test_ignore_one_line(self):
    self._create("""
    # line 2
    x = 123  # type: ignore
    # line 4
    """)
    self._should_report(True, 2)
    self._should_report(False, 3)
    self._should_report(True, 4)

  def test_ignore_one_line_mypy_style(self):
    self._create("""
    # line 2
    x = 123  # type: ignore[arg-type]
    # line 4
    """)
    self._should_report(True, 2)
    self._should_report(False, 3)
    self._should_report(True, 4)

  def test_utf8(self):
    self._create("""
    x = u"abc□def\\n"
    """)

  def test_ignore_extra_characters(self):
    self._create("""
    # line 2
    x = 123  # # type: ignore
    # line 4
    """)
    self._should_report(True, 2)
    self._should_report(False, 3)
    self._should_report(True, 4)

  def test_ignore_until_end(self):
    self._create("""
    # line 2
    # type: ignore
    # line 4
    """)
    self._should_report(True, 2)
    self._should_report(False, 3)
    self._should_report(False, 4)

  def test_out_of_scope(self):
    self._create("""
    # type: ignore
    """)
    self._should_report(False, 2)
    self._should_report(True, 2, filename=None)  # No file.
    self._should_report(True, 2, filename="some_other_file.py")  # Other file.
    self._should_report(False, None)  # No line number.
    self._should_report(False, 0)  # line number 0.

  def test_disable(self):
    self._create("""
    # line 2
    x = 123  # pytype: disable=test-error
    # line 4
    """)
    self._should_report(True, 2)
    self._should_report(False, 3)
    self._should_report(True, 4)

  def test_disable_extra_characters(self):
    self._create("""
    # line 2
    x = 123  # # pytype: disable=test-error
    # line 4
    """)
    self._should_report(True, 2)
    self._should_report(False, 3)
    self._should_report(True, 4)

  def test_disable_until_end(self):
    self._create("""
    # line 2
    # pytype: disable=test-error
    # line 4
    """)
    self._should_report(True, 2)
    self._should_report(False, 3)
    self._should_report(False, 4)

  def test_enable_after_disable(self):
    self._create("""
    # line 2
    # pytype: disable=test-error
    # line 4
    # pytype: enable=test-error
    """)
    self._should_report(True, 2)
    self._should_report(False, 3)
    self._should_report(False, 4)
    self._should_report(True, 5)
    self._should_report(True, 100)

  def test_enable_one_line(self):
    self._create("""
    # line 2
    # pytype: disable=test-error
    # line 4
    x = 123 # pytype: enable=test-error
    """)
    self._should_report(True, 2)
    self._should_report(False, 3)
    self._should_report(False, 4)
    self._should_report(True, 5)
    self._should_report(False, 6)
    self._should_report(False, 100)

  def test_disable_other_error(self):
    self._create("""
    # line 2
    x = 123  # pytype: disable=test-other-error
    # line 4
    """)
    self._should_report(True, 2)
    self._should_report(True, 3)
    self._should_report(False, 3, error_name="test-other-error")
    self._should_report(True, 4)

  def test_disable_multiple_error(self):
    self._create("""
    # line 2
    x = 123  # pytype: disable=test-error,test-other-error
    # line 4
    """)
    self._should_report(True, 2)
    self._should_report(False, 3)
    self._should_report(False, 3, error_name="test-other-error")
    self._should_report(True, 4)

  def test_disable_all(self):
    self._create("""
    # line 2
    x = 123  # pytype: disable=*
    # line 4
    """)
    self._should_report(True, 2)
    self._should_report(False, 3)
    self._should_report(True, 4)

  def test_multiple_directives(self):
    self._create("""
    x = 123  # sometool: directive=whatever # pytype: disable=test-error
    """)
    self._should_report(False, 2)

  def test_error_at_line_0(self):
    self._create("""
    x = "foo"
    # pytype: disable=attribute-error
    """)
    self._should_report(False, 0, error_name="attribute-error")

  def test_disable_without_space(self):
    self._create("""
    # line 2
    x = 123  # pytype:disable=test-error
    # line 4
    """)
    self._should_report(True, 2)
    self._should_report(False, 3)
    self._should_report(True, 4)

  def test_invalid_disable(self):
    def check_warning(message_regex, text):
      self._create(text)
      self.assertLessEqual(1, len(self._errorlog))
      error = list(self._errorlog)[0]
      self.assertEqual(_TEST_FILENAME, error._filename)
      self.assertEqual(1, error.lineno)
      self.assertRegex(str(error), message_regex)

    check_warning("Unknown pytype directive.*disalbe.*",
                  "# pytype: disalbe=test-error")
    check_warning("Invalid error name.*bad-error-name.*",
                  "# pytype: disable=bad-error-name")
    check_warning("Invalid directive syntax",
                  "# pytype: disable")
    check_warning("Invalid directive syntax",
                  "# pytype: ")
    check_warning("Unknown pytype directive.*foo.*",
                  "# pytype: disable=test-error foo=bar")
    # Spaces aren't allowed in the comma-separated value list.
    check_warning("Invalid directive syntax",
                  "# pytype: disable=test-error ,test-other-error")
    # This will actually result in two warnings: the first because the
    # empty string isn't a valid error name, the second because
    # test-other-error isn't a valid command.  We only verify the first
    # warning.
    check_warning("Invalid error name",
                  "# pytype: disable=test-error, test-other-error")

  def test_type_comments(self):
    self._create("""
    x = None  # type: int
    y = None  # allow extra comments # type: str
    z = None  # type: int  # and extra comments after, too
    a = None  # type:int  # without a space
    # type: (int, float) -> str
    # comment with embedded # type: should-be-discarded
    """)
    self.assertEqual({
        2: "int",
        3: "str",
        4: "int",
        5: "int",
        6: "(int, float) -> str",
    }, self._director.type_comments)

  def test_strings_that_look_like_directives(self):
    # Line 2 is a string, not a type comment.
    # Line 4 has a string and a comment.
    self._create("""
    s = "# type: int"
    x = None  # type: float
    y = "# type: int"  # type: str
    """)
    self.assertEqual({
        3: "float",
        4: "str",
    }, self._director.type_comments)

  def test_huge_string(self):
    # Tests that the director doesn't choke on this huge, repetitive file.
    src = ["x = ("]
    for i in range(2000):
      src.append(f"    'string{i}'")
    src.append(")")
    self._create("\n".join(src))

  def test_try(self):
    self._create("""
      try:
        x = None  # type: int
      except Exception:
        x = None  # type: str
      else:
        x = None  # type: float
    """)
    self.assertEqual({
        3: "int",
        5: "str",
        7: "float",
    }, self._director.type_comments)


class VariableAnnotationsTest(DirectorTestCase):

  def test_annotations(self):
    self._create("""
      v1: int = 0
      def f():
        v2: str = ''
    """)
    self.assertEqual({2: "int", 4: "str"}, self._director.annotations)

  def test_precedence(self):
    self._create("v: int = 0  # type: str")
    # Variable annotations take precedence. vm.py's _FindIgnoredTypeComments
    # warns about the ignored comment.
    self.assertEqual({1: "int"}, self._director.annotations)

  def test_parameter_annotation(self):
    # director.annotations contains only variable annotations and function type
    # comments, so a parameter annotation should be ignored.
    self._create("""
      def f(
          x: int = 0):
        pass
    """)
    self.assertFalse(self._director.annotations)

  def test_multistatement_line(self):
    self._create("""
      if __random__: v1: int = 0
      else: v2: str = ''
    """)
    self.assertEqual({2: "int", 3: "str"}, self._director.annotations)

  def test_multistatement_line_no_annotation(self):
    self._create("""
      if __random__: v = 0
      else: v = 1
    """)
    self.assertFalse(self._director.annotations)

  def test_comment_is_not_an_annotation(self):
    self._create("# FOMO(b/xxx): pylint: disable=invalid-name")
    self.assertFalse(self._director.annotations)

  def test_string_is_not_an_annotation(self):
    self._create("""
      logging.info('%s: completed: response=%s',  s1, s2)
      f(':memory:', bar=baz)
    """)
    self.assertFalse(self._director.annotations)

  def test_multiline_annotation(self):
    self._create("""
      v: Callable[  # a very important comment
          [], int] = None
    """)
    lineno = 2 if sys.version_info[:2] >= (3, 8) else 3
    self.assertEqual({lineno: "Callable[[], int]"}, self._director.annotations)

  def test_multiline_assignment(self):
    self._create("""
      v: List[int] = [
          0,
          1,
      ]
    """)
    lineno = 2 if sys.version_info[:2] >= (3, 8) else 4
    self.assertEqual({lineno: "List[int]"}, self._director.annotations)

  def test_complicated_annotation(self):
    self._create("v: int if __random__ else str = None")
    self.assertEqual(
        {1: "int if __random__ else str"}, self._director.annotations)

  def test_colon_in_value(self):
    self._create("v: Dict[str, int] = {x: y}")
    self.assertEqual({1: "Dict[str, int]"}, self._director.annotations)

  def test_equals_sign_in_value(self):
    self._create("v = {x: f(y=0)}")
    self.assertFalse(self._director.annotations)

  def test_annotation_after_comment(self):
    self._create("""
      # comment
      v: int = 0
    """)
    self.assertEqual({3: "int"}, self._director.annotations)


class LineNumbersTest(DirectorTestCase):

  def test_type_comment_on_multiline_value(self):
    self._create("""
      v = [
        ("hello",
         "world",  # type: should_be_ignored

        )
      ]  # type: dict
    """)
    # The line number of STORE_NAME v changes between versions.
    if self.python_version >= (3, 8):
      lineno = 2
    else:
      lineno = 3
    self.assertEqual({lineno: "dict"}, self._director.type_comments)

  def test_type_comment_with_trailing_comma(self):
    self._create("""
      v = [
        ("hello",
         "world"
        ),
      ]  # type: dict
      w = [
        ["hello",
         "world"
        ],  # some comment
      ]  # type: dict
    """)
    # The line numbers of STORE_NAME change between versions.
    if self.python_version >= (3, 8):
      v_lineno = 2
      w_lineno = 7
    else:
      v_lineno = 3
      w_lineno = 9
    self.assertEqual({
        v_lineno: "dict",
        w_lineno: "dict"
    }, self._director.type_comments)

  def test_decorators(self):
    self._create("""
      class A:
        '''
        @decorator in a docstring
        '''
        @real_decorator
        def f(x):
          x = foo @ bar @ baz

        @decorator(
            x, y
        )

        def bar():
          pass
    """)
    if self.python_version >= (3, 8):
      real_decorator_lineno = 7
      decorator_lineno = 14
    else:
      real_decorator_lineno = 6
      decorator_lineno = 11
    self.assertEqual(self._director.decorators,
                     {real_decorator_lineno, decorator_lineno})

  def test_stacked_decorators(self):
    self._create("""
      @decorator(
          x, y
      )

      @foo

      class A:
          pass
    """)
    lineno = 8 if self.python_version >= (3, 8) else 6
    self.assertEqual(self._director.decorators, {lineno})

  def test_overload(self):
    self._create("""
      from typing import overload

      @overload
      def f() -> int: ...

      @overload
      def f(x: str) -> str: ...

      def f(x=None):
        return 0 if x is None else x
    """)
    self.assertEqual(self._director.decorators, {5, 8})


class DisableDirectivesTest(DirectorTestCase):

  def assertDisables(self, *disable_lines, error_class=None, disables=None):
    assert not (error_class and disables)
    error_class = error_class or "wrong-arg-types"
    disables = disables or self._director._disables[error_class]
    for i in range(self.num_lines):
      lineno = i+1
      if lineno in disable_lines:
        self.assertIn(lineno, disables)
      else:
        self.assertNotIn(lineno, disables)

  def test_basic(self):
    self._create("""
      toplevel(
          a, b, c, d)  # pytype: disable=wrong-arg-types
    """)
    if self.python_version >= (3, 8):
      self.assertDisables(2, 3)
    else:
      self.assertDisables(3)

  def test_nested(self):
    self._create("""
      toplevel(
          nested())  # pytype: disable=wrong-arg-types
    """)
    if self.python_version >= (3, 8):
      self.assertDisables(2, 3)
    else:
      self.assertDisables(3)

  def test_multiple_nested(self):
    self._create("""
      toplevel(
        nested1(),
        nested2(),  # pytype: disable=wrong-arg-types
        nested3())
    """)
    if self.python_version >= (3, 8):
      self.assertDisables(2, 4)
    else:
      self.assertDisables(4)

  def test_multiple_toplevel(self):
    self._create("""
      toplevel1()
      toplevel2()  # pytype: disable=wrong-arg-types
      toplevel3()
    """)
    self.assertDisables(3)

  def test_deeply_nested(self):
    self._create("""
      toplevel(
        nested1(),
        nested2(
          deeply_nested1(),  # pytype: disable=wrong-arg-types
          deeply_nested2()),
        nested3())
    """)
    if self.python_version >= (3, 8):
      self.assertDisables(2, 4, 5)
    else:
      self.assertDisables(5)

  def test_non_toplevel(self):
    self._create("""
      x = [
        f("oops")  # pytype: disable=wrong-arg-types
      ]
    """)
    if self.python_version >= (3, 8):
      self.assertDisables(2, 3)
    else:
      self.assertDisables(3)

  def test_non_toplevel_bad_annotation(self):
    self._create("""
      x: list[int] = [
        f(
            "oops")]  # pytype: disable=annotation-type-mismatch
    """)
    if self.python_version >= (3, 8):
      self.assertDisables(2, 4, error_class="annotation-type-mismatch")
    else:
      self.assertDisables(4, error_class="annotation-type-mismatch")

  def test_trailing_parenthesis(self):
    self._create("""
      toplevel(
          a, b, c, d,
      )  # pytype: disable=wrong-arg-types
    """)
    if self.python_version >= (3, 8):
      self.assertDisables(2, 4)
    else:
      self.assertDisables(3, 4)

  def test_multiple_bytecode_blocks(self):
    self._create("""
      def f():
        call(a, b, c, d)
      def g():
        call(a, b, c, d)  # pytype: disable=wrong-arg-types
    """)
    self.assertDisables(5)

  def test_compare(self):
    self._create("""
      import datetime
      def f(right: datetime.date):
        left = datetime.datetime(1, 1, 1, 1)
        return left < right  # pytype: disable=wrong-arg-types
    """)
    self.assertDisables(5)

  def test_nested_compare(self):
    self._create("""
      f(
        a,
        b,
        (c <
         d)  # pytype: disable=wrong-arg-types
      )
    """)
    if self.python_version >= (3, 8):
      self.assertDisables(2, 5, 6)
    else:
      self.assertDisables(6)

  def test_iterate(self):
    self._create("""
      class Foo:
        def __iter__(self, too, many, args):
          pass
      foo = Foo()
      for x in foo:  # pytype: disable=missing-parameter
        print(x)
    """)
    self.assertDisables(6, error_class="missing-parameter")

  def test_subscript(self):
    self._create("""
      class Foo:
        def __getitem__(self, too, many, args):
          pass
      x = Foo()
      x['X']  # pytype: disable=missing-parameter
    """)
    self.assertDisables(6, error_class="missing-parameter")

  def test_attrs(self):
    self._create("""
      import attr
      def converter(x):
        return []
      @attr.s
      class Foo:
        x = attr.ib(
          converter=converter, factory=list, type=dict[str, str]
        )  # pytype: disable=annotation-type-mismatch
    """)
    if self.python_version >= (3, 8):
      self.assertDisables(7, 9, error_class="annotation-type-mismatch")
    else:
      self.assertDisables(8, 9, error_class="annotation-type-mismatch")

  def test_return(self):
    self._create("""
       def f(x):
         return x
       def g() -> int:
         return f(
             "oops")  # pytype: disable=bad-return-type
    """)
    if self.python_version >= (3, 8):
      self.assertDisables(5, 6, error_class="bad-return-type")
    else:
      self.assertDisables(6, error_class="bad-return-type")

  def test_if(self):
    self._create("""
      if (__random__ and
          name_error and  # pytype: disable=name-error
          __random__):
        pass
    """)
    self.assertDisables(3, error_class="name-error")

  def test_unsupported(self):
    self._create("""
      x = [
        "something_unsupported"
      ]  # pytype: disable=not-supported-yet
    """)
    if self.python_version >= (3, 8):
      self.assertDisables(2, 4, error_class="not-supported-yet")
    else:
      self.assertDisables(3, 4, error_class="not-supported-yet")

  def test_range(self):
    self._create("""
      f(
        # pytype: disable=attribute-error
        a.nonsense,
        b.nonsense,
        # pytype: enable=attribute-error
      )
    """)
    self.assertDisables(3, 4, 5, error_class="attribute-error")

  def test_ignore(self):
    # We have no idea if the '# type: ignore' is for the list construction, the
    # function call, or the function argument, so we apply it to all of them.
    self._create("""
      x = [
        some_bad_function(
            "some bad arg")]  # type: ignore
    """)
    if self.python_version >= (3, 8):
      self.assertDisables(2, 3, 4, disables=self._director.ignore)
    else:
      self.assertDisables(4, disables=self._director.ignore)

  def test_ignore_range(self):
    self._create("""
      x = [
        # type: ignore
        "oops"
      ]
    """)
    self.assertDisables(3, 4, 5, disables=self._director.ignore)

  def test_with_and_backslash_continuation(self):
    self._create("""
      with foo("a",
               "b"), \\
           bar("c",
               "d"),  \\
           baz("e"):  # pytype: disable=wrong-arg-types
        pass
    """)
    if self.python_version >= (3, 8):
      self.assertDisables(2, 6)
    else:
      self.assertDisables(6)

  def test_not_instantiable(self):
    self._create("""
      x = [
        A(
      )]  # pytype: disable=not-instantiable
    """)
    if self.python_version >= (3, 8):
      self.assertDisables(2, 3, 4, error_class="not-instantiable")
    else:
      self.assertDisables(3, 4, error_class="not-instantiable")

  def test_unsupported_operands_in_call(self):
    self._create("""
      some_func(
        x < y)  # pytype: disable=unsupported-operands
    """)
    if self.python_version >= (3, 8):
      self.assertDisables(2, 3, error_class="unsupported-operands")
    else:
      self.assertDisables(3, error_class="unsupported-operands")

  def test_unsupported_operands_in_assignment(self):
    self._create("""
      x["wrong key type"] = (
        some_call(),
        "oops")  # pytype: disable=unsupported-operands
    """)
    if self.python_version >= (3, 8):
      self.assertDisables(2, 4, error_class="unsupported-operands")
    else:
      self.assertDisables(4, error_class="unsupported-operands")

  def test_header(self):
    self._create("""
      if (x == 0 and
          (0).nonsense and  # pytype: disable=attribute-error
          y == 0):
        pass
    """)
    if self.python_version >= (3, 8):
      self.assertDisables(2, 3, error_class="attribute-error")
    else:
      self.assertDisables(3, error_class="attribute-error")

  def test_try(self):
    self._create("""
      try:
        pass
      except NonsenseError:  # pytype: disable=name-error
        pass
    """)
    self.assertDisables(4, error_class="name-error")

  def test_classdef(self):
    self._create("""
      import abc
      class Foo:  # pytype: disable=ignored-abstractmethod
        @abc.abstractmethod
        def f(self): ...
    """)
    self.assertDisables(3, error_class="ignored-abstractmethod")

  def test_class_attribute(self):
    self._create("""
      class Foo:
        x: 0  # pytype: disable=invalid-annotation
    """)
    self.assertDisables(3, error_class="invalid-annotation")

  def test_nested_call_in_function_decorator(self):
    self._create("""
      @decorate(
        dict(
          k1=v(
            a, b, c),  # pytype: disable=wrong-arg-types
          k2=v2))
      def f():
        pass
    """)
    if self.python_version >= (3, 8):
      self.assertDisables(2, 3, 4, 5)
    else:
      self.assertDisables(5)

  def test_nested_call_in_class_decorator(self):
    self._create("""
      @decorate(
        dict(
          k1=v(
            a, b, c),  # pytype: disable=wrong-arg-types
          k2=v2))
      class C:
        pass
    """)
    if self.python_version >= (3, 8):
      self.assertDisables(2, 3, 4, 5)
    else:
      self.assertDisables(5)


if __name__ == "__main__":
  unittest.main()



"""Debugging helper functions."""

import collections
import contextlib
import inspect
import io
import logging
import re
import traceback

from pytype import utils
from pytype.typegraph import cfg_utils

import tabulate


def _ascii_tree(out, node, p1, p2, seen, get_children, get_description=None):
  """Draw a graph, starting at a given position.

  Args:
    out: A file-like object to write the ascii tree to.
    node: The node from where to draw.
    p1: The current prefix.
    p2: The upcoming prefix.
    seen: Nodes we have seen so far (as a set).
    get_children: The function to call to retrieve children.
    get_description: Optional. A function to call to describe a node.
  """
  children = list(get_children(node))
  text = get_description(node) if get_description else str(node)
  if node in seen:
    out.write(p1 + "[" + text + "]\n")
  else:
    out.write(p1 + text + "\n")
    seen.add(node)
    for i, c in enumerate(children):
      last = (i == len(children) - 1)
      out.write(p2 + "|\n")
      _ascii_tree(out, c, p2 + "+-", p2 + ("  " if last else "| "),
                  seen, get_children, get_description)


def ascii_tree(node, get_children, get_description=None):
  """Draw a graph, starting at a given position.

  Args:
    node: The node from where to draw.
    get_children: The function to call to retrieve children.
    get_description: Optional. A function to call to describe a node.

  Returns:
    A string.
  """
  out = io.StringIO()
  _ascii_tree(out, node, "", "", set(), get_children, get_description)
  return out.getvalue()


def prettyprint_binding(binding, indent_level=0):
  """Pretty print a binding with variable id and data."""
  indent = " " * indent_level
  if not binding:
    return indent + "<>"
  return "%s<v%d : %r>" % (indent, binding.variable.id, binding.data)


def prettyprint_binding_set(binding_set, indent_level=0, label=""):
  """Pretty print a set of bindings, with optional label."""
  indent = " " * indent_level
  start = f"{indent}{label}: {{"
  if not binding_set:
    return start + " }"
  return "\n".join(
      [start] +
      [prettyprint_binding(x, indent_level + 2) for x in binding_set] +
      [indent + "}"])


def prettyprint_binding_nested(binding, indent_level=0):
  """Pretty print a binding and its recursive contents."""
  indent = " " * indent_level
  if indent_level > 32:
    return indent + "-[ max recursion depth exceeded ]-\n"
  s = f"{indent}binding v{binding.variable.id}={binding.data!r}\n"
  other = ""
  for v in binding.variable.bindings:
    if v is not binding:
      other += f"{v.data!r} {[o.where for o in v.origins]} "
  if other:
    s += f"{indent}(other assignments: {other})\n"
  for origin in binding.origins:
    s += f"{indent}  at {origin.where}\n"
    for i, source_set in enumerate(origin.source_sets):
      for j, source in enumerate(source_set):
        s += prettyprint_binding_nested(source, indent_level + 4)
        if j < len(source_set)-1:
          s += f"{indent}    AND\n"
      if i < len(origin.source_sets)-1:
        s += f"{indent}  OR\n"
  return s


def prettyprint_cfg_node(node, decorate_after_node=0, full=False):
  """A reasonably compact representation of all the bindings at a node.

  Args:
    node: The node to prettyprint.
    decorate_after_node: Don't print bindings unless node_id > this.
    full: Print the full string representation of a binding's data

  Returns:
    A prettyprinted node.
  """
  if node.id <= decorate_after_node:
    return repr(node) + f" [{len(node.bindings)} bindings]"
  if full:
    name = lambda x: getattr(x, "name", str(x))
  else:
    name = str
  bindings = collections.defaultdict(list)
  for b in node.bindings:
    bindings[b.variable.id].append(name(b.data))
  b = ", ".join(["%d:%s" % (k, "|".join(v))
                 for k, v in sorted(bindings.items())])
  return repr(node) + " [" + b + "]"


def prettyprint_cfg_tree(root, decorate_after_node=0, full=False,
                         forward=False):
  """Pretty print a cfg tree with the bindings at each node.

  Args:
    root: The root node.
    decorate_after_node: Don't print bindings unless node_id > this.
    full: Print the full string representation of a binding's data
    forward: Traverse the tree forwards if true.

  Returns:
    A prettyprinted tree.
  """
  if forward:
    children = lambda node: node.outgoing
  else:
    children = lambda node: node.incoming
  desc = lambda node: prettyprint_cfg_node(node, decorate_after_node, full)
  return ascii_tree(root, get_children=children, get_description=desc)


def _pretty_variable(var):
  """Return a pretty printed string for a Variable."""
  lines = []
  single_value = len(var.bindings) == 1
  var_desc = "v%d" % var.id
  if not single_value:
    # Write a description of the variable (for single value variables this
    # will be written along with the value later on).
    lines.append(var_desc)
    var_prefix = "  "
  else:
    var_prefix = var_desc + " = "

  for value in var.bindings:
    data = utils.maybe_truncate(value.data)
    binding = f"{var_prefix} {data}"

    if len(value.origins) == 1:
      # Single origin.  Use the binding as a prefix when writing the origin.
      prefix = binding + ", "
    else:
      # Multiple origins, write the binding on its own line, then indent all
      # of the origins.
      lines.append(binding)
      prefix = "    "

    for origin in value.origins:
      src = utils.pretty_dnf([[str(v) for v in source_set]
                              for source_set in origin.source_sets])
      lines.append("%s%s @%d" %(prefix, src, origin.where.id))
  return "\n".join(lines)


def program_to_text(program):
  """Generate a text (CFG nodes + assignments) version of a program.

  For debugging only.

  Args:
    program: An instance of cfg.Program

  Returns:
    A string representing all of the data for this program.
  """
  def label(node):
    return "<%d>%s" % (node.id, node.name)
  s = io.StringIO()
  seen = set()
  for node in cfg_utils.order_nodes(program.cfg_nodes):
    seen.add(node)
    s.write(f"{label(node)}\n")
    s.write(f"  From: {', '.join(label(n) for n in node.incoming)}\n")
    s.write(f"  To: {', '.join(label(n) for n in node.outgoing)}\n")
    s.write("\n")
    variables = {value.variable for value in node.bindings}
    for var in sorted(variables, key=lambda v: v.id):
      # If a variable is bound in more than one node then it will be listed
      # redundantly multiple times.  One alternative would be to only list the
      # values that occur in the given node, and then also list the other nodes
      # that assign the same variable.

      # Write the variable, indenting by two spaces.
      s.write("  %s\n" % _pretty_variable(var).replace("\n", "\n  "))
    s.write("\n")

  return s.getvalue()


def root_cause(binding, node, seen=()):
  """Tries to determine why a binding isn't possible at a node.

  This tries to find the innermost source that's still impossible. It only works
  if the failure isn't due to a combination of bindings.

  Args:
    binding: A binding, or a list of bindings.
    node: The node at which (one of the) binding(s) is impossible.
    seen: Internal. Bindings already looked at.

  Returns:
    A tuple (binding, node), with "binding" the innermost binding that's
    not possible, and "node" the CFG node at which it isn't.
  """
  if isinstance(binding, (list, tuple)):
    bindings = list(binding)
  else:
    bindings = [binding]
  del binding
  key = frozenset(bindings)
  if key in seen:
    return next(iter(bindings), None), node
  for b in bindings:
    if not node.HasCombination([b]):
      for o in b.origins:
        for source_set in o.source_sets:
          cause, n = root_cause(list(source_set), o.where)
          if cause is not None:
            return cause, n
      return b, node
  return None, None


def stack_trace(indent_level=0, limit=100):
  indent = " " * indent_level
  stack = [frame for frame in traceback.extract_stack()
           if "/errors.py" not in frame[0] and "/debug.py" not in frame[0]]
  tb = traceback.format_list(stack[-limit:])
  tb = [indent + re.sub(r"/usr/.*/pytype/", "", x) for x in tb]
  return "\n  ".join(tb)


def _setup_tabulate():
  """Customise tabulate."""
  tabulate.PRESERVE_WHITESPACE = True
  tabulate.MIN_PADDING = 0
  # Overwrite the 'presto' format to use the block-drawing vertical line.
  # pytype: disable=module-attr
  tabulate._table_formats["presto"] = tabulate.TableFormat(  # pylint: disable=protected-access
      lineabove=None,
      linebelowheader=tabulate.Line("", "-", "+", ""),
      linebetweenrows=None,
      linebelow=None,
      headerrow=tabulate.DataRow("", "│", ""),
      datarow=tabulate.DataRow("", "│", ""),
      padding=1, with_header_hide=None)
  # pytype: enable=module-attr


def show_ordered_code(code, extra_col=None):
  """Print out the block structure of an OrderedCode object as a table.

  Args:
    code: A blocks.OrderedCode object
    extra_col: A map from opcode_index to a single additional cell to display
  """
  if not extra_col:
    extra_col = {}
  _setup_tabulate()
  block_lines = []
  op_lines = []
  boundaries = []
  start = 0
  for block in code.order:
    end = start
    ids = lambda xs: [x.id for x in xs]
    block_lines.append(
        f"block: {block.id} -> {ids(block.outgoing)} <- {ids(block.incoming)}")
    for op in block:
      end += 1
      op_lines.append([
          op.index,
          op.__class__.__name__,
          getattr(op, "pretty_arg", ""),
          op.target and op.target.index,
          op.block_target and op.block_target.index,
          op.next and op.next.index,
          op.line,
          extra_col.get(op.index)
      ])
    boundaries.append((start, end))
    start = end
  headers = ["ix", "op", "arg", "tgt", "btgt", "next", "line", "extra"]
  block_table = tabulate.tabulate(op_lines, headers, tablefmt="presto")
  block_table = block_table.split("\n")
  tab = [[block_table[0]]]
  block_table = block_table[2:]
  for blk, (start, end) in zip(block_lines, boundaries):
    tab.append([blk])
    tab.append(["\n".join(block_table[start:end])])
  print(tabulate.tabulate(tab, tablefmt="fancy_grid"))


# Tracing logger
def tracer(name=None):
  name = f"trace.{name}" if name else "trace"
  return logging.getLogger(name)


def set_trace_level(level):
  logging.getLogger("trace").setLevel(level)


@contextlib.contextmanager
def tracing(level=logging.DEBUG):
  log = logging.getLogger("trace")
  current_level = log.getEffectiveLevel()
  log.setLevel(level)
  try:
    yield
  finally:
    log.setLevel(current_level)


def trace(name, *trace_args):
  """Record args and return value for a function call.

  The trace is of the form
    function name: {
    function name: arg = value
    function name: arg = value
    ...
    function name: -> return
    function name: }

  This will let us write tools to pretty print the traces with indentation etc.

  Args:
    name: module name, usually `__name__`
    *trace_args: function arguments to log
  Returns:
    a decorator
  """
  def decorator(f):
    def wrapper(*args, **kwargs):
      t = tracer(name)
      if t.getEffectiveLevel() < logging.DEBUG:
        return f(*args, **kwargs)
      argspec = inspect.getfullargspec(f)
      t.debug("%s: {", f.__name__)
      for arg in trace_args:
        if isinstance(arg, int):
          argname = argspec.args[arg]
          val = args[arg]
        else:
          argname = arg
          val = kwargs[arg]
        t.debug("%s: %s = %s", f.__name__, argname, show(val))
      ret = f(*args, **kwargs)
      t.debug("%s: -> %s", f.__name__, show(ret))
      t.debug("%s: }", f.__name__)
      return ret
    return wrapper
  return decorator


def show(x):
  """Pretty print values for debugging."""
  typename = x.__class__.__name__
  if typename == "Variable":
    return f"{x!r} {x.data}"
  else:
    return f"{x!r} <{typename}>"



from logging import Logger

log: Logger

class SDKConfig:
    XRAY_ENABLED_KEY: str
    DISABLED_ENTITY_NAME: str
    @classmethod
    def sdk_enabled(cls): ...
    @classmethod
    def set_sdk_enabled(cls, value) -> None: ...



from collections import OrderedDict
from collections.abc import Mapping
from typing import Any, overload

@overload
def alphabetize_attributes(attrs: None) -> None: ...
@overload
def alphabetize_attributes(attrs: Mapping[Any, str]) -> OrderedDict[Any, str]: ...



from _typeshed import Self
from typing import Iterable, MutableSequence, TypeVar, overload

_T = TypeVar("_T")

class UserList(MutableSequence[_T]):
    data: list[_T]
    def insert(self, index: int, object: _T) -> None: ...
    @overload
    def __setitem__(self, i: int, o: _T) -> None: ...
    @overload
    def __setitem__(self, s: slice, o: Iterable[_T]) -> None: ...
    def __delitem__(self, i: int | slice) -> None: ...
    def __len__(self) -> int: ...
    @overload
    def __getitem__(self, i: int) -> _T: ...
    @overload
    def __getitem__(self: Self, s: slice) -> Self: ...
    def sort(self) -> None: ...



from typing import Any

from ._chart import ChartBase as ChartBase

class ScatterChart(ChartBase):
    tagname: str
    scatterStyle: Any
    varyColors: Any
    ser: Any
    dLbls: Any
    dataLabels: Any
    extLst: Any
    x_axis: Any
    y_axis: Any
    __elements__: Any
    def __init__(
        self,
        scatterStyle: Any | None = ...,
        varyColors: Any | None = ...,
        ser=...,
        dLbls: Any | None = ...,
        extLst: Any | None = ...,
        **kw,
    ) -> None: ...



from typing import Any, ClassVar

import passlib.utils.handlers as uh

class HexDigestHash(uh.StaticHandler):
    checksum_chars: ClassVar[str]
    supported: ClassVar[bool]

def create_hex_hash(digest, module=..., django_name: Any | None = ..., required: bool = ...): ...

hex_md4: Any
hex_md5: Any
hex_sha1: Any
hex_sha256: Any
hex_sha512: Any

class htdigest(uh.MinimalHandler):
    name: ClassVar[str]
    default_encoding: ClassVar[str]
    @classmethod
    def hash(cls, secret, user, realm, encoding: Any | None = ...): ...  # type: ignore[override]
    @classmethod
    def verify(cls, secret, hash, user, realm, encoding: str = ...): ...  # type: ignore[override]
    @classmethod
    def identify(cls, hash): ...
    @classmethod
    def genconfig(cls): ...
    @classmethod
    def genhash(cls, secret, config, user, realm, encoding: Any | None = ...): ...  # type: ignore[override]



from braintree.exceptions.unexpected_error import UnexpectedError as UnexpectedError

class TimeoutError(UnexpectedError): ...
class ConnectTimeoutError(TimeoutError): ...
class ReadTimeoutError(TimeoutError): ...



from _typeshed import Self
from types import CodeType
from typing import Any, Callable, Text, TypeVar
from typing_extensions import ParamSpec

def run(statement: str, filename: str | None = ..., sort: str | int = ...) -> None: ...
def runctx(
    statement: str, globals: dict[str, Any], locals: dict[str, Any], filename: str | None = ..., sort: str | int = ...
) -> None: ...

_T = TypeVar("_T")
_P = ParamSpec("_P")
_Label = tuple[str, int, str]

class Profile:
    stats: dict[_Label, tuple[int, int, int, int, dict[_Label, tuple[int, int, int, int]]]]  # undocumented
    def __init__(
        self, timer: Callable[[], float] = ..., timeunit: float = ..., subcalls: bool = ..., builtins: bool = ...
    ) -> None: ...
    def enable(self) -> None: ...
    def disable(self) -> None: ...
    def print_stats(self, sort: str | int = ...) -> None: ...
    def dump_stats(self, file: Text) -> None: ...
    def create_stats(self) -> None: ...
    def snapshot_stats(self) -> None: ...
    def run(self: Self, cmd: str) -> Self: ...
    def runctx(self: Self, cmd: str, globals: dict[str, Any], locals: dict[str, Any]) -> Self: ...
    def runcall(self, __func: Callable[_P, _T], *args: _P.args, **kw: _P.kwargs) -> _T: ...

def label(code: str | CodeType) -> _Label: ...  # undocumented



from typing import Any

from ...engine.result import FilterResult

class AsyncCommon(FilterResult):
    async def close(self) -> None: ...

class AsyncResult(AsyncCommon):
    def __init__(self, real_result) -> None: ...
    def keys(self): ...
    def unique(self, strategy: Any | None = ...): ...
    def columns(self, *col_expressions): ...
    async def partitions(self, size: Any | None = ...) -> None: ...
    async def fetchone(self): ...
    async def fetchmany(self, size: Any | None = ...): ...
    async def all(self): ...
    def __aiter__(self): ...
    async def __anext__(self): ...
    async def first(self): ...
    async def one_or_none(self): ...
    async def scalar_one(self): ...
    async def scalar_one_or_none(self): ...
    async def one(self): ...
    async def scalar(self): ...
    async def freeze(self): ...
    def merge(self, *others): ...
    def scalars(self, index: int = ...): ...
    def mappings(self): ...

class AsyncScalarResult(AsyncCommon):
    def __init__(self, real_result, index) -> None: ...
    def unique(self, strategy: Any | None = ...): ...
    async def partitions(self, size: Any | None = ...) -> None: ...
    async def fetchall(self): ...
    async def fetchmany(self, size: Any | None = ...): ...
    async def all(self): ...
    def __aiter__(self): ...
    async def __anext__(self): ...
    async def first(self): ...
    async def one_or_none(self): ...
    async def one(self): ...

class AsyncMappingResult(AsyncCommon):
    def __init__(self, result) -> None: ...
    def keys(self): ...
    def unique(self, strategy: Any | None = ...): ...
    def columns(self, *col_expressions): ...
    async def partitions(self, size: Any | None = ...) -> None: ...
    async def fetchall(self): ...
    async def fetchone(self): ...
    async def fetchmany(self, size: Any | None = ...): ...
    async def all(self): ...
    def __aiter__(self): ...
    async def __anext__(self): ...
    async def first(self): ...
    async def one_or_none(self): ...
    async def one(self): ...



"""Tests for inspect.graph."""

from pytype.inspect import graph
from pytype.typegraph import cfg

import unittest


class GraphTest(unittest.TestCase):

  def setUp(self):
    super().setUp()
    self.prog = cfg.Program()
    self.current_location = self.prog.NewCFGNode()

  def test_program_to_dot(self):
    v1 = self.prog.NewVariable()
    b = v1.AddBinding("x", [], self.current_location)
    n = self.current_location.ConnectNew()
    v2 = self.prog.NewVariable()
    v2.AddBinding("y", {b}, n)
    # smoke test
    tg = graph.TypeGraph(self.prog, set(), False)
    assert isinstance(tg.to_dot(), str)


if __name__ == "__main__":
  unittest.main()



from datetime import datetime, timedelta
from typing import Any as tAny

class Any:
    type_url: tAny = ...
    value: tAny = ...
    def Pack(self, msg: tAny, type_url_prefix: str = ..., deterministic: tAny | None = ...) -> None: ...
    def Unpack(self, msg: tAny): ...
    def TypeName(self): ...
    def Is(self, descriptor: tAny): ...

class Timestamp:
    def ToJsonString(self) -> str: ...
    seconds: int = ...
    nanos: int = ...
    def FromJsonString(self, value: str) -> None: ...
    def GetCurrentTime(self) -> None: ...
    def ToNanoseconds(self) -> int: ...
    def ToMicroseconds(self) -> int: ...
    def ToMilliseconds(self) -> int: ...
    def ToSeconds(self) -> int: ...
    def FromNanoseconds(self, nanos: int) -> None: ...
    def FromMicroseconds(self, micros: int) -> None: ...
    def FromMilliseconds(self, millis: int) -> None: ...
    def FromSeconds(self, seconds: int) -> None: ...
    def ToDatetime(self) -> datetime: ...
    def FromDatetime(self, dt: datetime) -> None: ...

class Duration:
    def ToJsonString(self) -> str: ...
    seconds: int = ...
    nanos: int = ...
    def FromJsonString(self, value: tAny) -> None: ...
    def ToNanoseconds(self) -> int: ...
    def ToMicroseconds(self) -> int: ...
    def ToMilliseconds(self) -> int: ...
    def ToSeconds(self) -> int: ...
    def FromNanoseconds(self, nanos: int) -> None: ...
    def FromMicroseconds(self, micros: int) -> None: ...
    def FromMilliseconds(self, millis: int) -> None: ...
    def FromSeconds(self, seconds: int) -> None: ...
    def ToTimedelta(self) -> timedelta: ...
    def FromTimedelta(self, td: timedelta) -> None: ...

class FieldMask:
    def ToJsonString(self) -> str: ...
    def FromJsonString(self, value: tAny) -> None: ...
    def IsValidForDescriptor(self, message_descriptor: tAny): ...
    def AllFieldsFromDescriptor(self, message_descriptor: tAny) -> None: ...
    def CanonicalFormFromMask(self, mask: tAny) -> None: ...
    def Union(self, mask1: tAny, mask2: tAny) -> None: ...
    def Intersect(self, mask1: tAny, mask2: tAny) -> None: ...
    def MergeMessage(
        self, source: tAny, destination: tAny, replace_message_field: bool = ..., replace_repeated_field: bool = ...
    ) -> None: ...

class _FieldMaskTree:
    def __init__(self, field_mask: tAny | None = ...) -> None: ...
    def MergeFromFieldMask(self, field_mask: tAny) -> None: ...
    def AddPath(self, path: tAny): ...
    def ToFieldMask(self, field_mask: tAny) -> None: ...
    def IntersectPath(self, path: tAny, intersection: tAny): ...
    def AddLeafNodes(self, prefix: tAny, node: tAny) -> None: ...
    def MergeMessage(self, source: tAny, destination: tAny, replace_message: tAny, replace_repeated: tAny) -> None: ...

class Struct:
    def __getitem__(self, key: tAny): ...
    def __contains__(self, item: tAny): ...
    def __setitem__(self, key: tAny, value: tAny) -> None: ...
    def __delitem__(self, key: tAny) -> None: ...
    def __len__(self): ...
    def __iter__(self): ...
    def keys(self): ...
    def values(self): ...
    def items(self): ...
    def get_or_create_list(self, key: tAny): ...
    def get_or_create_struct(self, key: tAny): ...
    def update(self, dictionary: tAny) -> None: ...

class ListValue:
    def __len__(self): ...
    def append(self, value: tAny) -> None: ...
    def extend(self, elem_seq: tAny) -> None: ...
    def __getitem__(self, index: tAny): ...
    def __setitem__(self, index: tAny, value: tAny) -> None: ...
    def __delitem__(self, key: tAny) -> None: ...
    def items(self) -> None: ...
    def add_struct(self): ...
    def add_list(self): ...

WKTBASES: dict[str, type[tAny]]



from .propagation import (
    Format as Format,
    InvalidCarrierException as InvalidCarrierException,
    SpanContextCorruptedException as SpanContextCorruptedException,
    UnsupportedFormatException as UnsupportedFormatException,
)
from .scope import Scope as Scope
from .scope_manager import ScopeManager as ScopeManager
from .span import Span as Span, SpanContext as SpanContext
from .tracer import (
    Reference as Reference,
    ReferenceType as ReferenceType,
    Tracer as Tracer,
    child_of as child_of,
    follows_from as follows_from,
    start_child_span as start_child_span,
)

tracer: Tracer
is_tracer_registered: bool

def global_tracer() -> Tracer: ...
def set_global_tracer(value: Tracer) -> None: ...
def is_global_tracer_registered() -> bool: ...



from distutils.config import PyPIRCCommand
from typing import ClassVar

class upload(PyPIRCCommand):
    description: ClassVar[str]
    boolean_options: ClassVar[list[str]]
    def run(self) -> None: ...
    def upload_file(self, command, pyversion, filename) -> None: ...



from types import FrameType
from typing import Callable

SIG_DFL: int
SIG_IGN: int

ITIMER_REAL: int
ITIMER_VIRTUAL: int
ITIMER_PROF: int

NSIG: int

SIGABRT: int
SIGALRM: int
SIGBREAK: int  # Windows
SIGBUS: int
SIGCHLD: int
SIGCLD: int
SIGCONT: int
SIGEMT: int
SIGFPE: int
SIGHUP: int
SIGILL: int
SIGINFO: int
SIGINT: int
SIGIO: int
SIGIOT: int
SIGKILL: int
SIGPIPE: int
SIGPOLL: int
SIGPROF: int
SIGPWR: int
SIGQUIT: int
SIGRTMAX: int
SIGRTMIN: int
SIGSEGV: int
SIGSTOP: int
SIGSYS: int
SIGTERM: int
SIGTRAP: int
SIGTSTP: int
SIGTTIN: int
SIGTTOU: int
SIGURG: int
SIGUSR1: int
SIGUSR2: int
SIGVTALRM: int
SIGWINCH: int
SIGXCPU: int
SIGXFSZ: int

# Windows
CTRL_C_EVENT: int
CTRL_BREAK_EVENT: int

class ItimerError(IOError): ...

_HANDLER = Callable[[int, FrameType], None] | int | None

def alarm(time: int) -> int: ...
def getsignal(signalnum: int) -> _HANDLER: ...
def pause() -> None: ...
def setitimer(which: int, seconds: float, interval: float = ...) -> tuple[float, float]: ...
def getitimer(which: int) -> tuple[float, float]: ...
def set_wakeup_fd(fd: int) -> int: ...
def siginterrupt(signalnum: int, flag: bool) -> None: ...
def signal(signalnum: int, handler: _HANDLER) -> _HANDLER: ...
def default_int_handler(signum: int, frame: FrameType) -> None: ...



from typing import Any

def format_unicode(raw_value): ...
def format_integer(raw_value): ...
def format_binary(raw_value): ...
def format_uuid(raw_value): ...
def format_uuid_le(raw_value): ...
def format_boolean(raw_value): ...
def format_ad_timestamp(raw_value): ...

time_format: Any

def format_time(raw_value): ...
def format_ad_timedelta(raw_value): ...
def format_time_with_0_year(raw_value): ...
def format_sid(raw_value): ...



"""Tests for abstract_utils.py."""

from pytype import config
from pytype.abstract import abstract_utils
from pytype.tests import test_base
from pytype.tests import test_utils

import unittest


class GetViewsTest(test_base.UnitTest):

  def setUp(self):
    super().setUp()
    options = config.Options.create(python_version=self.python_version)
    self._ctx = test_utils.make_context(options)

  def test_basic(self):
    v1 = self._ctx.program.NewVariable([self._ctx.convert.unsolvable], [],
                                       self._ctx.root_node)
    v2 = self._ctx.program.NewVariable(
        [self._ctx.convert.int_type, self._ctx.convert.str_type], [],
        self._ctx.root_node)
    views = list(abstract_utils.get_views([v1, v2], self._ctx.root_node))
    self.assertCountEqual([{v1: views[0][v1], v2: views[0][v2]},
                           {v1: views[1][v1], v2: views[1][v2]}],
                          [{v1: v1.bindings[0], v2: v2.bindings[0]},
                           {v1: v1.bindings[0], v2: v2.bindings[1]}])

  def _test_optimized(self, skip_future_value, expected_num_views):
    v1 = self._ctx.program.NewVariable([self._ctx.convert.unsolvable], [],
                                       self._ctx.root_node)
    v2 = self._ctx.program.NewVariable(
        [self._ctx.convert.int_type, self._ctx.convert.str_type], [],
        self._ctx.root_node)
    views = abstract_utils.get_views([v1, v2], self._ctx.root_node)
    skip_future = None
    # To count the number of views. Doesn't matter what we put in here, as long
    # as it's one per view.
    view_markers = []
    while True:
      try:
        view = views.send(skip_future)
      except StopIteration:
        break
      # Accesses v1 only, so the v2 bindings should be deduplicated when
      # `skip_future` is True.
      view_markers.append(view[v1])
      skip_future = skip_future_value
    self.assertEqual(len(view_markers), expected_num_views)

  def test_skip(self):
    self._test_optimized(skip_future_value=True, expected_num_views=1)

  def test_no_skip(self):
    self._test_optimized(skip_future_value=False, expected_num_views=2)


if __name__ == "__main__":
  unittest.main()



import sys

class _Feature:
    def __init__(self, optionalRelease: sys._version_info, mandatoryRelease: sys._version_info, compiler_flag: int) -> None: ...
    def getOptionalRelease(self) -> sys._version_info: ...
    def getMandatoryRelease(self) -> sys._version_info: ...
    compiler_flag: int

absolute_import: _Feature
division: _Feature
generators: _Feature
nested_scopes: _Feature
print_function: _Feature
unicode_literals: _Feature
with_statement: _Feature
barry_as_FLUFL: _Feature
generator_stop: _Feature

if sys.version_info >= (3, 7):
    annotations: _Feature

all_feature_names: list[str]  # undocumented

if sys.version_info >= (3, 7):
    __all__ = [
        "all_feature_names",
        "absolute_import",
        "division",
        "generators",
        "nested_scopes",
        "print_function",
        "unicode_literals",
        "with_statement",
        "barry_as_FLUFL",
        "generator_stop",
        "annotations",
    ]
else:
    __all__ = [
        "all_feature_names",
        "absolute_import",
        "division",
        "generators",
        "nested_scopes",
        "print_function",
        "unicode_literals",
        "with_statement",
        "barry_as_FLUFL",
        "generator_stop",
    ]



from typing import Any

from .context import SpanContext

class Propagator:
    def inject(self, span_context: SpanContext, carrier: dict[Any, Any]) -> None: ...
    def extract(self, carrier: dict[Any, Any]) -> SpanContext: ...



import sys
from _typeshed import StrPath
from typing import Any, Protocol

if sys.version_info >= (3, 7):
    from py_compile import PycInvalidationMode

__all__ = ["compile_dir", "compile_file", "compile_path"]

class _SupportsSearch(Protocol):
    def search(self, string: str) -> Any: ...

if sys.version_info >= (3, 9):
    def compile_dir(
        dir: StrPath,
        maxlevels: int | None = ...,
        ddir: StrPath | None = ...,
        force: bool = ...,
        rx: _SupportsSearch | None = ...,
        quiet: int = ...,
        legacy: bool = ...,
        optimize: int = ...,
        workers: int = ...,
        invalidation_mode: PycInvalidationMode | None = ...,
        *,
        stripdir: str | None = ...,  # TODO: change to StrPath | None once https://bugs.python.org/issue40447 is resolved
        prependdir: StrPath | None = ...,
        limit_sl_dest: StrPath | None = ...,
        hardlink_dupes: bool = ...,
    ) -> int: ...
    def compile_file(
        fullname: StrPath,
        ddir: StrPath | None = ...,
        force: bool = ...,
        rx: _SupportsSearch | None = ...,
        quiet: int = ...,
        legacy: bool = ...,
        optimize: int = ...,
        invalidation_mode: PycInvalidationMode | None = ...,
        *,
        stripdir: str | None = ...,  # TODO: change to StrPath | None once https://bugs.python.org/issue40447 is resolved
        prependdir: StrPath | None = ...,
        limit_sl_dest: StrPath | None = ...,
        hardlink_dupes: bool = ...,
    ) -> int: ...

elif sys.version_info >= (3, 7):
    def compile_dir(
        dir: StrPath,
        maxlevels: int = ...,
        ddir: StrPath | None = ...,
        force: bool = ...,
        rx: _SupportsSearch | None = ...,
        quiet: int = ...,
        legacy: bool = ...,
        optimize: int = ...,
        workers: int = ...,
        invalidation_mode: PycInvalidationMode | None = ...,
    ) -> int: ...
    def compile_file(
        fullname: StrPath,
        ddir: StrPath | None = ...,
        force: bool = ...,
        rx: _SupportsSearch | None = ...,
        quiet: int = ...,
        legacy: bool = ...,
        optimize: int = ...,
        invalidation_mode: PycInvalidationMode | None = ...,
    ) -> int: ...

else:
    def compile_dir(
        dir: StrPath,
        maxlevels: int = ...,
        ddir: StrPath | None = ...,
        force: bool = ...,
        rx: _SupportsSearch | None = ...,
        quiet: int = ...,
        legacy: bool = ...,
        optimize: int = ...,
        workers: int = ...,
    ) -> int: ...
    def compile_file(
        fullname: StrPath,
        ddir: StrPath | None = ...,
        force: bool = ...,
        rx: _SupportsSearch | None = ...,
        quiet: int = ...,
        legacy: bool = ...,
        optimize: int = ...,
    ) -> int: ...

if sys.version_info >= (3, 7):
    def compile_path(
        skip_curdir: bool = ...,
        maxlevels: int = ...,
        force: bool = ...,
        quiet: int = ...,
        legacy: bool = ...,
        optimize: int = ...,
        invalidation_mode: PycInvalidationMode | None = ...,
    ) -> int: ...

else:
    def compile_path(
        skip_curdir: bool = ...,
        maxlevels: int = ...,
        force: bool = ...,
        quiet: int = ...,
        legacy: bool = ...,
        optimize: int = ...,
    ) -> int: ...



import optparse
from collections.abc import Iterable
from typing import Any
from typing_extensions import Literal

from docutils import ApplicationError
from docutils.io import FileOutput
from docutils.nodes import document

_list = list

class DependencyList:
    list: _list[str]
    file: FileOutput | None
    def __init__(self, output_file: str | None = ..., dependencies: Iterable[str] = ...) -> None: ...
    def set_output(self, output_file: str | None) -> None: ...
    def add(self, *filenames: str) -> None: ...
    def close(self) -> None: ...

_SystemMessageLevel = Literal[0, 1, 2, 3, 4]

class Reporter:
    DEBUG_LEVEL: Literal[0]
    INFO_LEVEL: Literal[1]
    WARNING_LEVEL: Literal[2]
    ERROR_LEVEL: Literal[3]
    SEVERE_LEVEL: Literal[4]

    source: str
    report_level: _SystemMessageLevel
    halt_level: _SystemMessageLevel
    def __getattr__(self, __name: str) -> Any: ...  # incomplete

class SystemMessage(ApplicationError):
    level: _SystemMessageLevel
    def __init__(self, system_message: object, level: _SystemMessageLevel): ...

def new_reporter(source_path: str, settings: optparse.Values) -> Reporter: ...
def new_document(source_path: str, settings: optparse.Values | None = ...) -> document: ...
def __getattr__(name: str) -> Any: ...  # incomplete



from collections.abc import Generator
from typing import Any

COORD_RE: Any
COL_RANGE: str
ROW_RANGE: str
RANGE_EXPR: str
ABSOLUTE_RE: Any
SHEET_TITLE: str
SHEETRANGE_RE: Any

def get_column_interval(start, end): ...
def coordinate_from_string(coord_string): ...
def absolute_coordinate(coord_string): ...

col: Any

def get_column_letter(idx): ...
def column_index_from_string(str_col): ...
def range_boundaries(range_string): ...
def rows_from_range(range_string) -> Generator[Any, None, None]: ...
def cols_from_range(range_string) -> Generator[Any, None, None]: ...
def coordinate_to_tuple(coordinate): ...
def range_to_tuple(range_string): ...
def quote_sheetname(sheetname): ...



from datetime import datetime
from gzip import GzipFile
from httplib import HTTPConnection, HTTPResponse, HTTPSConnection
from ssl import SSLContext
from StringIO import StringIO
from time import struct_time
from types import InstanceType
from typing import IO, Any, AnyStr, Callable, Iterable, Mapping, MutableMapping, Union

_Unmarshaller = Any
_timeTuple = tuple[int, int, int, int, int, int, int, int, int]
# Represents types that can be compared against a DateTime object
_dateTimeComp = unicode | DateTime | datetime
# A "host description" used by Transport factories
_hostDesc = Union[str, tuple[str, Mapping[Any, Any]]]

def escape(s: AnyStr, replace: Callable[[AnyStr, AnyStr, AnyStr], AnyStr] = ...) -> AnyStr: ...

MAXINT: int
MININT: int
PARSE_ERROR: int
SERVER_ERROR: int
APPLICATION_ERROR: int
SYSTEM_ERROR: int
TRANSPORT_ERROR: int
NOT_WELLFORMED_ERROR: int
UNSUPPORTED_ENCODING: int
INVALID_ENCODING_CHAR: int
INVALID_XMLRPC: int
METHOD_NOT_FOUND: int
INVALID_METHOD_PARAMS: int
INTERNAL_ERROR: int

class Error(Exception): ...

class ProtocolError(Error):
    url: str
    errcode: int
    errmsg: str
    headers: Any
    def __init__(self, url: str, errcode: int, errmsg: str, headers: Any) -> None: ...

class ResponseError(Error): ...

class Fault(Error):
    faultCode: Any
    faultString: str
    def __init__(self, faultCode: Any, faultString: str, **extra: Any) -> None: ...

boolean: type[bool]
Boolean: type[bool]

class DateTime:
    value: str
    def __init__(self, value: str | unicode | datetime | float | int | _timeTuple | struct_time = ...) -> None: ...
    def make_comparable(self, other: _dateTimeComp) -> tuple[unicode, unicode]: ...
    def __lt__(self, other: _dateTimeComp) -> bool: ...
    def __le__(self, other: _dateTimeComp) -> bool: ...
    def __gt__(self, other: _dateTimeComp) -> bool: ...
    def __ge__(self, other: _dateTimeComp) -> bool: ...
    def __eq__(self, other: _dateTimeComp) -> bool: ...  # type: ignore[override]
    def __ne__(self, other: _dateTimeComp) -> bool: ...  # type: ignore[override]
    def timetuple(self) -> struct_time: ...
    def __cmp__(self, other: _dateTimeComp) -> int: ...
    def decode(self, data: Any) -> None: ...
    def encode(self, out: IO[str]) -> None: ...

class Binary:
    data: str
    def __init__(self, data: str | None = ...) -> None: ...
    def __cmp__(self, other: Any) -> int: ...
    def decode(self, data: str) -> None: ...
    def encode(self, out: IO[str]) -> None: ...

WRAPPERS: tuple[type[Any], ...]

# Still part of the public API, but see http://bugs.python.org/issue1773632
FastParser: None
FastUnmarshaller: None
FastMarshaller: None

# xmlrpclib.py will leave ExpatParser undefined if it can't import expat from
# xml.parsers. Because this is Python 2.7, the import will succeed.
class ExpatParser:
    def __init__(self, target: _Unmarshaller) -> None: ...
    def feed(self, data: str): ...
    def close(self): ...

# TODO: Add xmllib.XMLParser as base class
class SlowParser:
    handle_xml: Callable[[str, bool], None]
    unknown_starttag: Callable[[str, Any], None]
    handle_data: Callable[[str], None]
    handle_cdata: Callable[[str], None]
    unknown_endtag: Callable[[str, Callable[[Iterable[str], str], str]], None]
    def __init__(self, target: _Unmarshaller) -> None: ...

class Marshaller:
    memo: MutableMapping[int, Any]
    data: str | None
    encoding: str | None
    allow_none: bool
    def __init__(self, encoding: str | None = ..., allow_none: bool = ...) -> None: ...
    dispatch: Mapping[type, Callable[[Marshaller, str, Callable[[str], None]], None]]
    def dumps(
        self,
        values: Iterable[
            None
            | int
            | bool
            | long
            | float
            | str
            | unicode
            | list[Any]
            | tuple[Any, ...]
            | Mapping[Any, Any]
            | datetime
            | InstanceType
        ]
        | Fault,
    ) -> str: ...
    def dump_nil(self, value: None, write: Callable[[str], None]) -> None: ...
    def dump_int(self, value: int, write: Callable[[str], None]) -> None: ...
    def dump_bool(self, value: bool, write: Callable[[str], None]) -> None: ...
    def dump_long(self, value: long, write: Callable[[str], None]) -> None: ...
    def dump_double(self, value: float, write: Callable[[str], None]) -> None: ...
    def dump_string(
        self,
        value: str,
        write: Callable[[str], None],
        escape: Callable[[AnyStr, Callable[[AnyStr, AnyStr, AnyStr], AnyStr]], AnyStr] = ...,
    ) -> None: ...
    def dump_unicode(
        self,
        value: unicode,
        write: Callable[[str], None],
        escape: Callable[[AnyStr, Callable[[AnyStr, AnyStr, AnyStr], AnyStr]], AnyStr] = ...,
    ) -> None: ...
    def dump_array(self, value: Iterable[Any], write: Callable[[str], None]) -> None: ...
    def dump_struct(
        self,
        value: Mapping[unicode, Any],
        write: Callable[[str], None],
        escape: Callable[[AnyStr, Callable[[AnyStr, AnyStr, AnyStr], AnyStr]], AnyStr] = ...,
    ) -> None: ...
    def dump_datetime(self, value: datetime, write: Callable[[str], None]) -> None: ...
    def dump_instance(self, value: InstanceType, write: Callable[[str], None]) -> None: ...

class Unmarshaller:
    def append(self, object: Any) -> None: ...
    def __init__(self, use_datetime: bool = ...) -> None: ...
    def close(self) -> tuple[Any, ...]: ...
    def getmethodname(self) -> str | None: ...
    def xml(self, encoding: str, standalone: bool) -> None: ...
    def start(self, tag: str, attrs: Any) -> None: ...
    def data(self, text: str) -> None: ...
    def end(self, tag: str, join: Callable[[Iterable[str], str], str] = ...) -> None: ...
    def end_dispatch(self, tag: str, data: str) -> None: ...
    dispatch: Mapping[str, Callable[[Unmarshaller, str], None]]
    def end_nil(self, data: str): ...
    def end_boolean(self, data: str) -> None: ...
    def end_int(self, data: str) -> None: ...
    def end_double(self, data: str) -> None: ...
    def end_string(self, data: str) -> None: ...
    def end_array(self, data: str) -> None: ...
    def end_struct(self, data: str) -> None: ...
    def end_base64(self, data: str) -> None: ...
    def end_dateTime(self, data: str) -> None: ...
    def end_value(self, data: str) -> None: ...
    def end_params(self, data: str) -> None: ...
    def end_fault(self, data: str) -> None: ...
    def end_methodName(self, data: str) -> None: ...

class _MultiCallMethod:
    def __init__(self, call_list: list[tuple[str, tuple[Any, ...]]], name: str) -> None: ...

class MultiCallIterator:
    def __init__(self, results: list[Any]) -> None: ...

class MultiCall:
    def __init__(self, server: ServerProxy) -> None: ...
    def __getattr__(self, name: str) -> _MultiCallMethod: ...
    def __call__(self) -> MultiCallIterator: ...

def getparser(use_datetime: bool = ...) -> tuple[ExpatParser | SlowParser, Unmarshaller]: ...
def dumps(
    params: tuple[Any, ...] | Fault,
    methodname: str | None = ...,
    methodresponse: bool | None = ...,
    encoding: str | None = ...,
    allow_none: bool = ...,
) -> str: ...
def loads(data: str, use_datetime: bool = ...) -> tuple[tuple[Any, ...], str | None]: ...
def gzip_encode(data: str) -> str: ...
def gzip_decode(data: str, max_decode: int = ...) -> str: ...

class GzipDecodedResponse(GzipFile):
    stringio: StringIO[Any]
    def __init__(self, response: HTTPResponse) -> None: ...
    def close(self): ...

class _Method:
    def __init__(self, send: Callable[[str, tuple[Any, ...]], Any], name: str) -> None: ...
    def __getattr__(self, name: str) -> _Method: ...
    def __call__(self, *args: Any) -> Any: ...

class Transport:
    user_agent: str
    accept_gzip_encoding: bool
    encode_threshold: int | None
    def __init__(self, use_datetime: bool = ...) -> None: ...
    def request(self, host: _hostDesc, handler: str, request_body: str, verbose: bool = ...) -> tuple[Any, ...]: ...
    verbose: bool
    def single_request(self, host: _hostDesc, handler: str, request_body: str, verbose: bool = ...) -> tuple[Any, ...]: ...
    def getparser(self) -> tuple[ExpatParser | SlowParser, Unmarshaller]: ...
    def get_host_info(self, host: _hostDesc) -> tuple[str, list[tuple[str, str]] | None, Mapping[Any, Any] | None]: ...
    def make_connection(self, host: _hostDesc) -> HTTPConnection: ...
    def close(self) -> None: ...
    def send_request(self, connection: HTTPConnection, handler: str, request_body: str) -> None: ...
    def send_host(self, connection: HTTPConnection, host: str) -> None: ...
    def send_user_agent(self, connection: HTTPConnection) -> None: ...
    def send_content(self, connection: HTTPConnection, request_body: str) -> None: ...
    def parse_response(self, response: HTTPResponse) -> tuple[Any, ...]: ...

class SafeTransport(Transport):
    def __init__(self, use_datetime: bool = ..., context: SSLContext | None = ...) -> None: ...
    def make_connection(self, host: _hostDesc) -> HTTPSConnection: ...

class ServerProxy:
    def __init__(
        self,
        uri: str,
        transport: Transport | None = ...,
        encoding: str | None = ...,
        verbose: bool = ...,
        allow_none: bool = ...,
        use_datetime: bool = ...,
        context: SSLContext | None = ...,
    ) -> None: ...
    def __getattr__(self, name: str) -> _Method: ...
    def __call__(self, attr: str) -> Transport | None: ...

Server = ServerProxy



from typing import Any

from sqlalchemy.dialects.firebird.base import (
    BIGINT as BIGINT,
    BLOB as BLOB,
    CHAR as CHAR,
    DATE as DATE,
    FLOAT as FLOAT,
    NUMERIC as NUMERIC,
    SMALLINT as SMALLINT,
    TEXT as TEXT,
    TIME as TIME,
    TIMESTAMP as TIMESTAMP,
    VARCHAR as VARCHAR,
)

__all__ = (
    "SMALLINT",
    "BIGINT",
    "FLOAT",
    "FLOAT",
    "DATE",
    "TIME",
    "TEXT",
    "NUMERIC",
    "FLOAT",
    "TIMESTAMP",
    "VARCHAR",
    "CHAR",
    "BLOB",
    "dialect",
)

dialect: Any



"""Functions for computing the execution order of bytecode."""

from pytype.pyc import loadmarshal
from pytype.pyc import opcodes
from pytype.pyc import pyc
from pytype.typegraph import cfg_utils

STORE_OPCODES = (
    opcodes.STORE_NAME,
    opcodes.STORE_FAST,
    opcodes.STORE_ATTR,
    opcodes.STORE_DEREF,
    opcodes.STORE_GLOBAL)

# Opcodes whose argument can be a block of code.
CODE_LOADING_OPCODES = (opcodes.LOAD_CONST,)


class OrderedCode:
  """Code object which knows about instruction ordering.

  Attributes:
    co_argcount: Same as loadmarshal.CodeType.
    co_posonlyargcount: Same as loadmarshal.CodeType.
    co_kwonlyargcount: Same as loadmarshal.CodeType.
    co_nlocals: Same as loadmarshal.CodeType.
    co_stacksize: Same as loadmarshal.CodeType.
    co_flags: Same as loadmarshal.CodeType.
    co_consts: Same as loadmarshal.CodeType.
    co_names: Same as loadmarshal.CodeType.
    co_varnames: Same as loadmarshal.CodeType.
    co_filename: Same as loadmarshal.CodeType.
    co_name: Same as loadmarshal.CodeType.
    co_firstlineno: Same as loadmarshal.CodeType.
    co_lnotab: Same as loadmarshal.CodeType.
    co_freevars: Same as loadmarshal.CodeType.
    co_cellvars: Same as loadmarshal.CodeType.
    order: A list of bytecode blocks. They're ordered ancestors-first, see
      cfg_utils.py:order_nodes.
    code_iter: A flattened list of block opcodes. Corresponds to co_code.
    original_co_code: The original code object's co_code.
    first_opcode: The first opcode in code_iter.
    python_version: The Python version this bytecode is from.
  """

  _HAS_DYNAMIC_ATTRIBUTES = True

  def __init__(self, code, bytecode, order, python_version):
    # Copy all "co_*" attributes from code.
    # This is preferable to both inheritance (because we don't want to be
    # compatible with the base class, which is too low level) as well as
    # object composition (because that would make the API too clunky for
    # callers).
    # NOTE: We don't copy co_code; callers should use self.code_iter instead.
    assert hasattr(code, "co_code")
    self.__dict__.update({name: value for name, value in code.__dict__.items()
                          if name.startswith("co_") and name != "co_code"})
    self.order = order
    # Keep the original co_code around temporarily to work around an issue in
    # the block collection algorithm (b/191517403)
    self.original_co_code = bytecode
    self.python_version = python_version
    for insn in bytecode:
      insn.code = self

  @property
  def code_iter(self):
    return (op for block in self.order for op in block)  # pylint: disable=g-complex-comprehension

  @property
  def first_opcode(self):
    return next(self.code_iter)

  def has_opcode(self, op_type):
    return any(isinstance(op, op_type) for op in self.code_iter)

  def has_iterable_coroutine(self):
    return bool(self.co_flags & loadmarshal.CodeType.CO_ITERABLE_COROUTINE)

  def set_iterable_coroutine(self):
    self.co_flags |= loadmarshal.CodeType.CO_ITERABLE_COROUTINE

  def has_coroutine(self):
    return bool(self.co_flags & loadmarshal.CodeType.CO_COROUTINE)

  def has_generator(self):
    return bool(self.co_flags & loadmarshal.CodeType.CO_GENERATOR)

  def has_async_generator(self):
    return bool(self.co_flags & loadmarshal.CodeType.CO_ASYNC_GENERATOR)

  def has_varargs(self):
    return bool(self.co_flags & loadmarshal.CodeType.CO_VARARGS)

  def has_varkeywords(self):
    return bool(self.co_flags & loadmarshal.CodeType.CO_VARKEYWORDS)

  def has_newlocals(self):
    return bool(self.co_flags & loadmarshal.CodeType.CO_NEWLOCALS)

  def get_arg_count(self):
    count = self.co_argcount + max(self.co_kwonlyargcount, 0)
    if self.has_varargs():
      count += 1
    if self.has_varkeywords():
      count += 1
    return count


class Block:
  """A block is a node in a directed graph.

  It has incoming and outgoing edges (jumps). Incoming jumps always jump
  to the first instruction of our bytecode, and outgoing jumps always jump
  from the last instruction. There are no jump instructions in the middle of
  a byte code block.
  A block implements most of the "sequence" interface, i.e., it can be used as
  if it was a Python list of bytecode instructions.

  Attributes:
    id: Block id
    code: A bytecode object (a list of instances of opcodes.Opcode).
    incoming: Incoming edges. These are blocks that jump to the first
      instruction in our code object.
    outgoing: Outgoing edges. These are the targets jumped to by the last
      instruction in our code object.
  """

  def __init__(self, code):
    self.id = code[0].index
    self.code = code
    self.incoming = set()
    self.outgoing = set()

  def connect_outgoing(self, target):
    """Add an outgoing edge."""
    self.outgoing.add(target)
    target.incoming.add(self)

  def __str__(self):
    return "<Block %d>" % self.id

  def __repr__(self):
    return "<Block %d: %r>" % (self.id, self.code)

  def __getitem__(self, index_or_slice):
    return self.code.__getitem__(index_or_slice)

  def __iter__(self):
    return self.code.__iter__()


def add_pop_block_targets(bytecode, python_version):
  """Modifies bytecode so that each POP_BLOCK has a block_target.

  This is to achieve better initial ordering of try/except and try/finally code.
  try:
    i = 1
    a[i]
  except IndexError:
    return i
  By connecting a CFG edge from the end of the block (after the "a[i]") to the
  except handler, our basic block ordering algorithm knows that the except block
  needs to be scheduled last, whereas if there only was an edge before the
  "i = 1", it would be able to schedule it too early and thus encounter an
  undefined variable. This is only for ordering. The actual analysis of the
  code happens later, in vm.py.

  Args:
    bytecode: An array of bytecodes.
    python_version: The target python version.
  """
  if not bytecode:
    return

  for op in bytecode:
    op.block_target = None

  if python_version >= (3, 8):
    setup_except_op = opcodes.SETUP_FINALLY
  else:
    setup_except_op = opcodes.SETUP_EXCEPT
  todo = [(bytecode[0], ())]  # unordered queue of (position, block_stack)
  seen = set()
  while todo:
    op, block_stack = todo.pop()
    if op in seen:
      continue
    else:
      seen.add(op)

    # Compute the block stack
    if isinstance(op, opcodes.POP_BLOCK):
      assert block_stack, "POP_BLOCK without block."
      op.block_target = block_stack[-1].target
      block_stack = block_stack[0:-1]
    elif isinstance(op, opcodes.RAISE_VARARGS):
      # Make "raise" statements jump to the innermost exception handler.
      # (If there's no exception handler, do nothing.)
      for b in reversed(block_stack):
        if isinstance(b, setup_except_op):
          op.block_target = b.target
          break
    elif isinstance(op, opcodes.BREAK_LOOP):
      # Breaks jump to after the loop
      for i in reversed(range(len(block_stack))):
        b = block_stack[i]
        if isinstance(b, opcodes.SETUP_LOOP):
          op.block_target = b.target
          assert b.target != op
          todo.append((op.block_target, block_stack[0:i]))
          break
    elif isinstance(op, setup_except_op):
      # Exceptions pop the block, so store the previous block stack.
      todo.append((op.target, block_stack))
      block_stack += (op,)
    elif op.pushes_block():
      assert op.target, f"{op.name} without target"
      # We push the entire opcode onto the block stack, for better debugging.
      block_stack += (op,)
    elif op.does_jump() and op.target:
      todo.append((op.target, block_stack))

    if not op.no_next():
      assert op.next, f"Bad instruction at end of bytecode: {op!r}."
      todo.append((op.next, block_stack))


def _split_bytecode(bytecode):
  """Given a sequence of bytecodes, return basic blocks.

  This will split the code at "basic block boundaries". These occur at
  every instruction that is jumped to, and after every instruction that jumps
  somewhere else (or returns / aborts).

  Args:
    bytecode: A list of instances of opcodes.Opcode. (E.g. returned from
      opcodes.dis())

  Returns:
    A list of _Block instances.
  """
  targets = {op.target for op in bytecode if op.target}
  blocks = []
  code = []
  for op in bytecode:
    code.append(op)
    if (op.no_next() or op.does_jump() or op.pops_block() or
        op.next is None or op.next in targets):
      blocks.append(Block(code))
      code = []
  return blocks


def compute_order(bytecode):
  """Split bytecode into blocks and order the blocks.

  This builds an "ancestor first" ordering of the basic blocks of the bytecode.

  Args:
    bytecode: A list of instances of opcodes.Opcode. (E.g. returned from
      opcodes.dis())

  Returns:
    A list of Block instances.
  """
  blocks = _split_bytecode(bytecode)
  first_op_to_block = {block.code[0]: block for block in blocks}
  for i, block in enumerate(blocks):
    next_block = blocks[i + 1] if i < len(blocks) - 1 else None
    first_op, last_op = block.code[0], block.code[-1]
    if next_block and not last_op.no_next():
      block.connect_outgoing(next_block)
    if first_op.target:
      # Handles SETUP_EXCEPT -> except block
      block.connect_outgoing(first_op_to_block[first_op.target])
    if last_op.target:
      block.connect_outgoing(first_op_to_block[last_op.target])
    if last_op.block_target:
      block.connect_outgoing(first_op_to_block[last_op.block_target])
  return cfg_utils.order_nodes(blocks)


class DisCodeVisitor:
  """Visitor for disassembling code into Opcode objects."""

  def visit_code(self, code):
    code.co_code = opcodes.dis_code(code)
    return code


def order_code(code, python_version):
  """Split a CodeType object into ordered blocks.

  This takes a CodeType object (i.e., a piece of compiled Python code) and
  splits it into ordered basic blocks.

  Args:
    code: A loadmarshal.CodeType object.
    python_version: The target python version.

  Returns:
    A CodeBlocks instance.
  """
  bytecodes = code.co_code
  add_pop_block_targets(bytecodes, python_version)
  return OrderedCode(code, bytecodes, compute_order(bytecodes),
                     code.python_version)


class OrderCodeVisitor:
  """Visitor for recursively changing all CodeType to OrderedCode.

  Depends on DisCodeVisitor having been run first.
  """

  def __init__(self, python_version):
    self._python_version = python_version

  def visit_code(self, code):
    return order_code(code, self._python_version)


class CollectAnnotationTargetsVisitor:
  """Collect opcodes that might have annotations attached.

  Depends on DisCodeVisitor having been run first.
  """

  def __init__(self):
    # A mutable map of line: opcode for STORE_* opcodes. This is modified as the
    # visitor runs, and contains the last opcode for each line.
    self.store_ops = {}
    # A mutable map of start: (end, opcode) for MAKE_FUNCTION opcodes. This is
    # modified as the visitor runs, and contains the range of lines that could
    # contain function type comments.
    self.make_function_ops = {}

  def visit_code(self, code):
    """Find STORE_* and MAKE_FUNCTION opcodes for attaching annotations."""
    # Offset between function code and MAKE_FUNCTION
    # [LOAD_CONST <code>, LOAD_CONST name, MAKE_FUNCTION]
    offset = 2
    co_code = code.original_co_code
    for i, op in enumerate(co_code):
      if isinstance(op, opcodes.MAKE_FUNCTION):
        code_op = co_code[i - offset]
        assert isinstance(code_op, CODE_LOADING_OPCODES)
        fn_code = code.co_consts[code_op.arg]
        if not _is_function_def(fn_code):
          continue
        # First line of code in body.
        end_line = min(op.line for op in fn_code.original_co_code)
        self.make_function_ops[op.line] = (end_line, op)
      elif (isinstance(op, STORE_OPCODES) and
            op.line not in self.make_function_ops):
        # For type comments attached to multi-opcode lines, we want to mark the
        # latest 'store' opcode and attach the type comment to it.
        self.store_ops[op.line] = op
    return code


def _is_function_def(fn_code):
  """Helper function for CollectFunctionTypeCommentTargetsVisitor."""
  # Reject anything that is not a named function (e.g. <lambda>).
  first = fn_code.co_name[0]
  if not (first == "_" or first.isalpha()):
    return False

  # Class definitions generate a constructor function. We can distinguish them
  # by checking for code blocks that start with LOAD_NAME __name__
  op = fn_code.first_opcode
  if (isinstance(op, opcodes.LOAD_NAME) and
      op.pretty_arg == "__name__"):
    return False

  return True


def merge_annotations(code, annotations):
  """Merges type comments into their associated opcodes.

  Modifies code in place.

  Args:
    code: An OrderedCode object.
    annotations: A map of lines to annotations.

  Returns:
    The code with annotations added to the relevant opcodes.
  """
  visitor = CollectAnnotationTargetsVisitor()
  code = pyc.visit(code, visitor)

  # Apply type comments to the STORE_* opcodes
  for line, op in visitor.store_ops.items():
    if line in annotations:
      op.annotation = annotations[line]

  # Apply type comments to the MAKE_FUNCTION opcodes
  for start, (end, op) in sorted(
      visitor.make_function_ops.items(), reverse=True):
    for i in range(start, end):
      # Take the first comment we find as the function typecomment.
      if i in annotations:
        # Record the line number of the comment for error messages.
        op.annotation = (annotations[i], i)
        break
  return code


def process_code(code, python_version):
  # [binary opcodes] -> [pyc.Opcode]
  ops = pyc.visit(code, DisCodeVisitor())
  # pyc.load_marshal.CodeType -> blocks.OrderedCode
  ordered = pyc.visit(ops, OrderCodeVisitor(python_version))
  return ordered



"""Merges type annotations from pyi files into the corresponding py files."""

import difflib
import enum
import os
import shutil

from typing import List, Optional, Tuple

import libcst as cst
from libcst import codemod
from libcst.codemod import visitors


class MergeError(Exception):
  """Wrap exceptions thrown while merging files."""


def _merge_csts(*, py_tree, pyi_tree):
  context = codemod.CodemodContext()
  vis = visitors.ApplyTypeAnnotationsVisitor
  vis.store_stub_in_context(context, pyi_tree)
  return vis(
      context,
      strict_posargs_matching=False,
      strict_annotation_matching=True,
  ).transform_module(py_tree)


def merge_sources(*, py, pyi):
  try:
    py_cst = cst.parse_module(py)
    pyi_cst = cst.parse_module(pyi)
    merged_cst = _merge_csts(py_tree=py_cst, pyi_tree=pyi_cst)
    return merged_cst.code
  except Exception as e:  # pylint: disable=broad-except
    raise MergeError(str(e)) from e


class Mode(enum.Enum):
  PRINT = 1
  DIFF = 2
  OVERWRITE = 3


def _get_diff(a, b):
  a, b = a.split("\n"), b.split("\n")
  diff = difflib.Differ().compare(a, b)
  return "\n".join(diff)


def merge_files(
    *,
    py_path: str,
    pyi_path: str,
    mode: Mode,
    backup: Optional[str] = None
) -> bool:
  """Merges a .py and a .pyi file."""

  with open(py_path) as f:
    py_src = f.read()
  with open(pyi_path) as f:
    pyi_src = f.read()
  annotated_src = merge_sources(py=py_src, pyi=pyi_src)
  changed = annotated_src != py_src
  if mode == Mode.PRINT:
    # Always print to stdout even if we haven't changed anything.
    print(annotated_src)
  elif mode == Mode.DIFF and changed:
    diff = _get_diff(py_src, annotated_src)
    print(diff)
  elif mode == Mode.OVERWRITE and changed:
    if backup:
      shutil.copyfile(py_path, f"{py_path}.{backup}")
    with open(py_path, "w") as f:
      f.write(annotated_src)
  return changed


def merge_tree(
    *,
    py_path: str,
    pyi_path: str,
    backup: Optional[str] = None,
    verbose: bool = False
) -> Tuple[List[str], List[Tuple[str, MergeError]]]:

  """Merge .py files in a tree with the corresponding .pyi files."""

  errors = []
  changed_files = []

  for root, _, files in os.walk(py_path):
    rel = os.path.relpath(py_path, root)
    pyi_dir = os.path.normpath(os.path.join(pyi_path, rel))
    for f in files:
      if f.endswith(".py"):
        py = os.path.join(root, f)
        pyi = os.path.join(pyi_dir, f + "i")
        if os.path.exists(pyi):
          if verbose:
            print("Merging:", py, end=" ")
          try:
            changed = merge_files(
                py_path=py, pyi_path=pyi, mode=Mode.OVERWRITE, backup=backup)
            if changed:
              changed_files.append(py)
            if verbose:
              print("[OK]")
          except MergeError as e:
            errors.append((py, e))
            if verbose:
              print("[FAILED]")
  return changed_files, errors



"""Base support for generating classes from data declarations.

Contains common functionality used by dataclasses, attrs and namedtuples.
"""

import abc
import collections
import logging
from typing import Any, ClassVar, Dict

from pytype import special_builtins
from pytype.abstract import abstract
from pytype.abstract import abstract_utils
from pytype.abstract import class_mixin
from pytype.overlays import overlay_utils


log = logging.getLogger(__name__)


# type aliases for convenience
Param = overlay_utils.Param
Attribute = class_mixin.Attribute
AttributeKinds = class_mixin.AttributeKinds


# Probably should make this an enum.Enum at some point.
class Ordering:
  """Possible orderings for get_class_locals."""
  # Order by each variable's first annotation. For example, for
  #   class Foo:
  #     x: int
  #     y: str
  #     x: float
  # the locals will be [(x, Instance(float)), (y, Instance(str))]. Note that
  # unannotated variables will be skipped, and the values of later annotations
  # take precedence over earlier ones.
  FIRST_ANNOTATE = object()
  # Order by each variable's last definition. So for
  #   class Foo:
  #     x = 0
  #     y = 'hello'
  #     x = 4.2
  # the locals will be [(y, Instance(str)), (x, Instance(float))]. Note that
  # variables without assignments will be skipped.
  LAST_ASSIGN = object()


class Decorator(abstract.PyTDFunction, metaclass=abc.ABCMeta):
  """Base class for decorators that generate classes from data declarations."""

  # Defaults for the args that we support (dataclasses only support 'init',
  # but the others default to false so they should not affect anything).
  _DEFAULT_ARGS: ClassVar[Dict[str, Any]] = {
      "init": True,
      "kw_only": False,
      "auto_attribs": False,
  }

  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)
    # Decorator.call() is invoked first with args, then with the class to
    # decorate, so we need to first store the args and then associate them to
    # the right class.
    self._current_args = None
    # Some constructors like attr.dataclass partially apply args, overriding the
    # defaults attached to the class.
    self.partial_args = {}
    self.args = {}  # map from each class we decorate to its args

  @abc.abstractmethod
  def decorate(self, node, cls):
    """Apply the decorator to cls."""

  def get_initial_args(self):
    ret = self._DEFAULT_ARGS.copy()
    ret.update(self.partial_args)
    return ret

  def update_kwargs(self, args):
    """Update current_args with the Args passed to the decorator."""
    self._current_args = self.get_initial_args()
    for k, v in args.namedargs.items():
      if k in self._current_args:
        try:
          self._current_args[k] = abstract_utils.get_atomic_python_constant(v)
        except abstract_utils.ConversionError:
          self.ctx.errorlog.not_supported_yet(
              self.ctx.vm.frames, f"Non-constant argument to decorator: {k!r}")

  def set_current_args(self, kwargs):
    """Set current_args when constructing a class directly."""
    self._current_args = self.get_initial_args()
    self._current_args.update(kwargs)

  def init_name(self, attr):
    """Attribute name as an __init__ keyword, could differ from attr.name."""
    return attr.name

  def make_init(self, node, cls, attrs, init_method_name="__init__"):
    pos_params = []
    kwonly_params = []
    all_kwonly = self.args[cls]["kw_only"]
    for attr in attrs:
      if not attr.init:
        continue
      typ = attr.init_type or attr.typ
      # call self.init_name in case the name differs from the field name -
      # e.g. attrs removes leading underscores from attrib names when
      # generating kwargs for __init__.
      param = Param(
          name=self.init_name(attr),
          typ=typ,
          default=attr.default)

      # kw_only=False in a field does not override kw_only=True in the class.
      if all_kwonly or attr.kw_only:
        kwonly_params.append(param)
      else:
        pos_params.append(param)

    return overlay_utils.make_method(self.ctx, node, init_method_name,
                                     pos_params, 0, kwonly_params)

  def call(self, node, func, args):
    """Construct a decorator, and call it on the class."""
    args = args.simplify(node, self.ctx)
    self.match_args(node, args)

    # There are two ways to use a decorator:
    #   @decorator(...)
    #   class Foo: ...
    # or
    #   @decorator
    #   class Foo: ...
    # In the first case, call() is invoked twice: once with kwargs to create the
    # decorator object and once with the decorated class as a posarg. So we call
    # update_kwargs on the first invocation, setting _current_args, and skip it
    # on the second.
    # In the second case, we call update_kwargs on the first and only
    # invocation.
    if not self._current_args:
      self.update_kwargs(args)

    # NOTE: @dataclass is py3-only and has explicitly kwonly args in its
    # constructor.
    #
    # @attr.s does not take positional arguments in typical usage, but
    # technically this works:
    #   class Foo:
    #     x = attr.ib()
    #   Foo = attr.s(Foo, **kwargs)
    #
    # Unfortunately, it also works to pass kwargs as posargs; we will at least
    # reject posargs if the first arg is not a Callable.
    if not args.posargs:
      return node, self.to_variable(node)

    cls_var = args.posargs[0]
    # We should only have a single binding here
    cls, = cls_var.data

    if not isinstance(cls, abstract.Class):
      # There are other valid types like abstract.Unsolvable that we don't need
      # to do anything with.
      return node, cls_var

    self.args[cls] = self._current_args
    # Reset _current_args so we don't use old args for a new class.
    self._current_args = None

    # decorate() modifies the cls object in place
    self.decorate(node, cls)
    return node, cls_var


class FieldConstructor(abstract.PyTDFunction):
  """Implements constructors for fields."""

  def get_kwarg(self, args, name, default):
    if name not in args.namedargs:
      return default
    try:
      return abstract_utils.get_atomic_python_constant(args.namedargs[name])
    except abstract_utils.ConversionError:
      self.ctx.errorlog.not_supported_yet(self.ctx.vm.frames,
                                          f"Non-constant argument {name!r}")

  def get_positional_names(self):
    # TODO(mdemello): We currently assume all field constructors are called with
    # namedargs, which has worked in practice but is not required by the attrs
    # or dataclasses apis.
    return []


def is_method(var):
  if var is None:
    return False
  return isinstance(var.data[0], (
      abstract.INTERPRETER_FUNCTION_TYPES,
      special_builtins.ClassMethodInstance,
      special_builtins.PropertyInstance,
      special_builtins.StaticMethodInstance
  ))


def is_dunder(name):
  return name.startswith("__") and name.endswith("__")


def add_member(node, cls, name, typ):
  if typ.formal:
    # If typ contains a type parameter, we mark it as empty so that instances
    # will use __annotations__ to fill in concrete type parameter values.
    instance = typ.ctx.convert.empty.to_variable(node)
  else:
    # See test_attr.TestAttrib.test_repeated_default - keying on the name
    # prevents attributes from sharing the same default object.
    _, instance = typ.ctx.vm.init_class(node, typ, extra_key=name)
  cls.members[name] = instance


def is_relevant_class_local(class_local: abstract_utils.Local,
                            class_local_name: str, allow_methods: bool):
  """Tests whether the current class local could be relevant for type checking.

  For example, this doesn't match __dunder__ class locals.

  To get an abstract_utils.Local from a vm.LocalOps, you can use,
  'vm_instance.annotated_locals[cls_name][op.name]'.

  Args:
    class_local: the local to query
    class_local_name: the name of the class local (because abstract_utils.Local
      does not hold this information).
    allow_methods: whether to allow methods class locals to match

  Returns:
    Whether this class local could possibly be relevant for type checking.
      Callers will usually want to filter even further.
  """
  if is_dunder(class_local_name):
    return False
  if not allow_methods and not class_local.typ and is_method(class_local.orig):
    return False
  return True


def get_class_locals(cls_name: str, allow_methods: bool, ordering, ctx):
  """Gets a dictionary of the class's local variables.

  Args:
    cls_name: The name of an abstract.InterpreterClass.
    allow_methods: A bool, whether to allow methods as variables.
    ordering: A classgen.Ordering describing the order in which the variables
      should appear.
    ctx: The abstract context.

  Returns:
    A collections.OrderedDict of the locals.
  """
  out = collections.OrderedDict()
  if cls_name not in ctx.vm.local_ops:
    # See TestAttribPy3.test_cannot_decorate in tests/test_attr2.py. The
    # class will not be in local_ops if a previous decorator hides it.
    return out
  for op in ctx.vm.local_ops[cls_name]:
    local = ctx.vm.annotated_locals[cls_name][op.name]
    if not is_relevant_class_local(local, op.name, allow_methods):
      continue
    if ordering is Ordering.FIRST_ANNOTATE:
      if not op.is_annotate() or op.name in out:
        continue
    else:
      assert ordering is Ordering.LAST_ASSIGN
      if not op.is_assign():
        continue
      elif op.name in out:
        out.move_to_end(op.name)
    out[op.name] = local
  return out


def make_replace_method(ctx, node, cls, *, kwargs_name="kwargs"):
  """Create a replace() method for a dataclass."""
  # This is used by several packages that extend dataclass.
  # The signature is
  #   def replace(self: T, **kwargs) -> T
  typevar = abstract.TypeParameter(abstract_utils.T + cls.name, ctx, bound=cls)
  return overlay_utils.make_method(
      ctx=ctx,
      node=node,
      name="replace",
      return_type=typevar,
      self_param=overlay_utils.Param("self", typevar),
      kwargs=overlay_utils.Param(kwargs_name),
  )


def get_or_create_annotations_dict(members, ctx):
  """Get __annotations__ from members map, create and attach it if not present.

  The returned dict is also referenced by members, so it is safe to mutate.

  Args:
    members: A dict of member name to variable.
    ctx: context.Context instance.

  Returns:
    members['__annotations__'] unpacked as a python dict
  """
  annotations_dict = abstract_utils.get_annotations_dict(members)
  if annotations_dict is None:
    annotations_dict = abstract.AnnotationsDict({}, ctx)
    members["__annotations__"] = annotations_dict.to_variable(ctx.root_node)
  return annotations_dict



from collections.abc import Generator
from typing import Any

from openpyxl.descriptors.serialisable import Serialisable

mimetypes: Any

class FileExtension(Serialisable):
    tagname: str
    Extension: Any
    ContentType: Any
    def __init__(self, Extension, ContentType) -> None: ...

class Override(Serialisable):
    tagname: str
    PartName: Any
    ContentType: Any
    def __init__(self, PartName, ContentType) -> None: ...

DEFAULT_TYPES: Any
DEFAULT_OVERRIDE: Any

class Manifest(Serialisable):
    tagname: str
    Default: Any
    Override: Any
    path: str
    __elements__: Any
    def __init__(self, Default=..., Override=...) -> None: ...
    @property
    def filenames(self): ...
    @property
    def extensions(self): ...
    def to_tree(self): ...
    def __contains__(self, content_type): ...
    def find(self, content_type): ...
    def findall(self, content_type) -> Generator[Any, None, None]: ...
    def append(self, obj) -> None: ...
